from collections import ChainMap
from dataclasses import dataclass
from typing import Any, Union

import pysmt
import z3 as z3lib

Z3 = z3lib.z3
from pysmt import logics
from pysmt import shortcuts as SMT

from asdl_adt import ADT, validators
from asdl_adt.validators import ValidationError
from .LoopIR import T, LoopIR
from .prelude import *

_first_run = True


def _get_smt_solver():
    factory = pysmt.factory.Factory(pysmt.shortcuts.get_env())
    slvs = factory.all_solvers(logic=logics.LIA)
    if len(slvs) == 0:
        raise OSError("Could not find any SMT solvers")
    return pysmt.shortcuts.Solver(name=next(iter(slvs)))


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Analysis Expr


def is_type_bound(val):
    if isinstance(val, (tuple, LoopIR.type)):
        return val
    raise ValidationError(Union[tuple, LoopIR.type], type(val))


class AOp(str):
    front_ops = {
        "+",
        "-",
        "*",
        "/",
        "%",
        "<",
        ">",
        "<=",
        ">=",
        "==",
        "and",
        "or",
        "==>",
    }

    def __new__(cls, op):
        op = str(op)
        if op in AOp.front_ops:
            return super().__new__(cls, op)
        raise ValueError(f"invalid operator: {op}")


A = ADT(
    """
module AExpr {
    expr    = Var( sym name )
            | Unk() -- unknown
            | Not( expr arg )
            | USub( expr arg )
            | Const( object val )
            | ConstSym( sym name ) -- represents a named, opaque value
            | BinOp( binop op, expr lhs, expr rhs )
            | Stride( sym name, int dim )
            | LetStrides( sym name, expr* strides, expr body )
            | Select( expr cond, expr tcase, expr fcase )
            | ForAll( sym name, expr arg )
            | Exists( sym name, expr arg )
            | Definitely( expr arg )
            | Maybe( expr arg )
            | Tuple( expr* args )
            | LetTuple( sym* names, expr rhs, expr body )
            | Let( sym* names, expr* rhs, expr body )
            attributes( type type, srcinfo srcinfo )
} """,
    {
        "sym": Sym,
        "type": is_type_bound,
        "binop": validators.instance_of(AOp, convert=True),
        "srcinfo": SrcInfo,
    },
)

# constructor helpers...
def AInt(x):
    if type(x) is int:
        return A.Const(x, T.int, null_srcinfo())
    elif isinstance(x, Sym):
        return A.Var(x, T.index, null_srcinfo())
    else:
        assert False, f"bad type {type(x)}"


def ABool(x):
    if type(x) is bool:
        return A.Const(x, T.bool, null_srcinfo())
    elif isinstance(x, Sym):
        return A.Var(x, T.bool, null_srcinfo())
    else:
        assert False, f"bad type {type(x)}"


def ALet(names, rhs, body):
    assert isinstance(names, list) and isinstance(rhs, list)
    assert all(isinstance(n, Sym) for n in names)
    assert all(isinstance(r, A.expr) for r in rhs)
    if isinstance(body, A.Let):
        names = names + body.names
        rhs = rhs + body.rhs
        body = body.body
    if len(names) == 0:
        return body
    else:
        return A.Let(names, rhs, body, body.type, body.srcinfo)


def ALetStride(nm, strides, body):
    if len(strides) == 0:
        return body
    else:
        return A.LetStrides(nm, strides, body, body.type, body.srcinfo)


def ALetTuple(names, rhs, body):
    if len(names) == 0:
        return body
    else:
        return A.LetTuple(names, rhs, body, body.type, body.srcinfo)


def AForAll(names, body):
    for nm in reversed(names):
        body = A.ForAll(nm, body, T.bool, body.srcinfo)
    return body


def AExists(names, body):
    for nm in reversed(names):
        body = A.Exists(nm, body, T.bool, body.srcinfo)
    return body


def ANot(x):
    return A.Not(x, T.bool, x.srcinfo)


def AAnd(*args):
    if len(args) == 0:
        return A.Const(True, T.bool, null_srcinfo())
    res = args[0]
    for a in args[1:]:
        res = A.BinOp("and", res, a, T.bool, a.srcinfo)
    return res


def AOr(*args):
    if len(args) == 0:
        return A.Const(False, T.bool, null_srcinfo())
    res = args[0]
    for a in args[1:]:
        res = A.BinOp("or", res, a, T.bool, a.srcinfo)
    return res


def AImplies(lhs, rhs):
    return A.BinOp("==>", lhs, rhs, T.bool, lhs.srcinfo)


def AEq(lhs, rhs):
    return A.BinOp("==", lhs, rhs, T.bool, lhs.srcinfo)


def AMay(arg):
    return A.Maybe(arg, T.bool, arg.srcinfo)


def ADef(arg):
    return A.Definitely(arg, T.bool, arg.srcinfo)


@extclass(A.expr)
def __neg__(arg):
    return A.USub(arg, T.bool, arg.srcinfo)


# USub
# Binop
#   + - * / %  < > <= >= ==  and or
@extclass(A.expr)
def __add__(lhs, rhs):
    return A.BinOp("+", lhs, rhs, T.index, lhs.srcinfo)


@extclass(A.expr)
def __sub__(lhs, rhs):
    return A.BinOp("-", lhs, rhs, T.index, lhs.srcinfo)


@extclass(A.expr)
def __mul__(lhs, rhs):
    return A.BinOp("*", lhs, rhs, T.index, lhs.srcinfo)


@extclass(A.expr)
def __truediv__(lhs, rhs):
    return A.BinOp("/", lhs, rhs, T.index, lhs.srcinfo)


@extclass(A.expr)
def __mod__(lhs, rhs):
    return A.BinOp("%", lhs, rhs, T.index, lhs.srcinfo)


@extclass(A.expr)
def __lt__(lhs, rhs):
    return A.BinOp("<", lhs, rhs, T.bool, lhs.srcinfo)


@extclass(A.expr)
def __gt__(lhs, rhs):
    return A.BinOp(">", lhs, rhs, T.bool, lhs.srcinfo)


@extclass(A.expr)
def __le__(lhs, rhs):
    return A.BinOp("<=", lhs, rhs, T.bool, lhs.srcinfo)


@extclass(A.expr)
def __ge__(lhs, rhs):
    return A.BinOp(">=", lhs, rhs, T.bool, lhs.srcinfo)


op_prec = {
    "exists": 10,
    "forall": 10,
    "==>": 10,
    #
    "ternary": 20,
    #
    "or": 30,
    #
    "and": 40,
    #
    "<": 50,
    ">": 50,
    "<=": 50,
    ">=": 50,
    "==": 50,
    #
    "+": 60,
    "-": 60,
    #
    "*": 70,
    "/": 70,
    "%": 70,
    #
    "unary": 90,
}

binop_print = {
    "==>": "⇒",
    "or": "∨",
    "and": "∧",
    "<": "<",
    ">": ">",
    "<=": "≤",
    ">=": "≥",
    "==": "==",
    "+": "+",
    "-": "-",
    "*": "*",
    "/": "/",
    "%": "%",
}


def _estr(e, prec=0, tab=""):
    if isinstance(e, A.Var):
        return str(e.name)
    elif isinstance(e, A.Unk):
        return "⊥"
    elif isinstance(e, A.Not):
        return f"¬{_estr(e.arg,op_prec['unary'],tab=tab)}"
    elif isinstance(e, A.USub):
        return f"-{_estr(e.arg,op_prec['unary'],tab=tab)}"
    elif isinstance(e, A.Const):
        return str(e.val)
    elif isinstance(e, A.ConstSym):
        return f"CONST({e.name})"
    elif isinstance(e, A.BinOp):
        local_prec = op_prec[e.op]
        lhs = _estr(e.lhs, prec=local_prec, tab=tab)
        rhs = _estr(e.rhs, prec=local_prec + 1, tab=tab)
        if local_prec < prec:
            return f"({lhs} {binop_print[e.op]} {rhs})"
        else:
            return f"{lhs} {binop_print[e.op]} {rhs}"
    elif isinstance(e, A.Stride):
        return f"stride({e.name},{e.dim})"
    elif isinstance(e, A.LetStrides):
        strides = ",".join([_estr(s, tab=tab + "  ") for s in e.strides])
        bind = f"{e.name} = ({strides})"
        body = _estr(e.body, tab=tab + "  ")
        s = f"letStride {bind}\n{tab}in {body}"
        return f"({s}\n{tab})" if prec > 0 else s
    elif isinstance(e, A.Select):
        local_prec = op_prec["ternary"]
        cond = _estr(e.cond, tab=tab)
        tcase = _estr(e.tcase, prec=local_prec + 1, tab=tab)
        fcase = _estr(e.fcase, prec=local_prec + 1, tab=tab)
        if local_prec < prec:
            return f"(({cond})? {tcase} : {fcase})"
        else:
            return f"({cond})? {tcase} : {fcase}"
    elif isinstance(e, (A.ForAll, A.Exists)):
        op = "∀" if isinstance(e, A.ForAll) else "∃"
        local_prec = op_prec["forall" if isinstance(e, A.ForAll) else "exists"]
        s = f"{op}{e.name},{_estr(e.arg,op_prec['forall'],tab=tab)}"
        if local_prec < prec:
            s = f"({s})"
        return s
    elif isinstance(e, (A.Definitely, A.Maybe)):
        op = "D" if isinstance(e, A.Definitely) else "M"
        return f"{op}{_estr(e.arg,op_prec['unary'],tab=tab)}"
    elif isinstance(e, A.Let):
        # compress nested lets for printing
        if isinstance(e.body, A.Let):
            return _estr(
                A.Let(
                    e.names + e.body.names,
                    e.rhs + e.body.rhs,
                    e.body.body,
                    e.type,
                    e.srcinfo,
                ),
                prec=prec,
                tab=tab,
            )
        binds = "\n".join(
            [f"{tab}{x} = {_estr(rhs,tab=tab+'  ')}" for x, rhs in zip(e.names, e.rhs)]
        )
        body = _estr(e.body, tab=tab + "  ")
        s = f"let\n{binds}\n{tab}in {body}"
        return f"({s}\n{tab})" if prec > 0 else s
    elif isinstance(e, A.Tuple):
        args = ", ".join([_estr(a, tab=tab) for a in e.args])
        return f"({args})"
    elif isinstance(e, A.LetTuple):
        names = ",".join([str(n) for n in e.names])
        bind = f"{names} = {_estr(e.rhs,tab=tab+'  ')}"
        body = _estr(e.body, tab=tab + "  ")
        s = f"let_tuple {bind}\n{tab}in {body}"
        return f"({s}\n{tab})" if prec > 0 else s
    else:
        assert False, "bad case"


@extclass(A.expr)
def __str__(e):
    return _estr(e)


@dataclass
class ConstSymFV:
    name: Sym

    def __hash__(self):
        return self.name.__hash__()


def aeFV(e, env=None):
    env = env or ChainMap()

    def push():
        nonlocal env
        env = env.new_child()

    def pop():
        nonlocal env
        env = env.parents

    if isinstance(e, A.Var):
        if e.name not in env:
            return {e.name: e.type}
        else:
            return dict()
    elif isinstance(e, (A.Unk, A.Const)):
        return dict()
    elif isinstance(e, (A.Not, A.USub, A.Definitely, A.Maybe)):
        return aeFV(e.arg, env)
    elif isinstance(e, A.BinOp):
        return aeFV(e.lhs, env) | aeFV(e.rhs, env)
    elif isinstance(e, A.ConstSym):
        return {ConstSymFV(e.name): e.type}
    elif isinstance(e, A.Stride):
        # stride symbol gets encoded as a tuple
        key = (e.name, e.dim)
        if key not in env:
            return {key: T.stride}
        else:
            return dict()
    elif isinstance(e, A.LetStrides):
        push()
        res = dict()
        for s in e.strides:
            res = res | aeFV(s, env)
        for i, _ in enumerate(e.strides):
            env[(e.name, i)] = True
        res = res | aeFV(e.body, env)
        pop()
        return res
    elif isinstance(e, A.Select):
        return aeFV(e.cond, env) | aeFV(e.tcase, env) | aeFV(e.fcase, env)
    elif isinstance(e, (A.ForAll, A.Exists)):
        push()
        env[e.name] = True
        res = aeFV(e.arg, env)
        pop()
        return res
    elif isinstance(e, A.Let):
        push()
        res = dict()
        for r in e.rhs:
            res = res | aeFV(r, env)
        for nm in e.names:
            env[nm] = True
        res = res | aeFV(e.body, env)
        pop()
        return res
    elif isinstance(e, A.Tuple):
        res = dict()
        for a in e.args:
            res = res | aeFV(a, env)
        return res
    elif isinstance(e, A.LetTuple):
        push()
        res = dict()
        res = aeFV(e.rhs, env)
        for nm in e.names:
            env[nm] = True
        res = res | aeFV(e.body, env)
        pop()
        return res
    else:
        assert False, "bad case"


def aeNegPos(e, pos, env=None, res=None):
    res = res or dict()
    env = env or dict()  # ChainMap()

    def save_and_shadow(names):
        nonlocal env
        return {nm: env.pop(nm, None) for nm in names}

    def restore_shadowed(saved):
        nonlocal env
        for nm, save_val in saved.items():
            if save_val is not None:
                env[nm] = save_val
            elif nm in env:
                del env[nm]

    # set the result for this node independent of cases
    res[id(e)] = pos

    if isinstance(e, A.Var):
        # backwards propagate through variable references
        if e.name not in env:
            env[e.name] = pos
        else:
            old_pos = env[e.name]
            if pos != old_pos:
                env[e.name] = "0"
    elif isinstance(e, A.Stride):
        # stride symbol gets encoded as a tuple
        key = (e.name, e.dim)
        if key not in env:
            env[key] = pos
        else:
            old_pos = env[e.name]
            if pos != old_pos:
                env[e.name] = "0"
    elif isinstance(e, (A.Unk, A.Const, A.Stride, A.ConstSym)):
        pass
    elif isinstance(e, A.Not):
        negpos = "-" if pos == "+" else "+" if pos == "-" else "0"
        aeNegPos(e.arg, negpos, env, res)
    elif isinstance(e, (A.USub, A.Definitely, A.Maybe)):
        aeNegPos(e.arg, pos, env, res)
    elif isinstance(e, A.BinOp):
        aeNegPos(e.lhs, pos, env, res)
        aeNegPos(e.rhs, pos, env, res)
    elif isinstance(e, A.LetStrides):
        keys = [(e.name, i) for i, _ in enumerate(e.strides)]
        # propagate negation-position flags through body
        saved = save_and_shadow(keys)
        aeNegPos(e.body, pos, env, res)
        key_pos = {k: env.get(k, pos) for k in keys}
        restore_shadowed(saved)

        # propagate values pushed onto strides in the body back to
        # their definitions
        for k, s, kp in zip(keys, e.strides, key_pos):
            aeNegPos(s, kp, env, res)

    elif isinstance(e, A.Select):
        aeNegPos(e.cond, pos, env, res)
        aeNegPos(e.tcase, pos, env, res)
        aeNegPos(e.fcase, pos, env, res)
    elif isinstance(e, (A.ForAll, A.Exists)):
        saved = save_and_shadow([e.name])
        aeNegPos(e.arg, pos, env, res)
        restore_shadowed(saved)
    elif isinstance(e, A.Let):
        save_stack = [save_and_shadow([nm]) for nm in e.names]
        aeNegPos(e.body, pos, env, res)
        for nm, rhs, save in reversed(list(zip(e.names, e.rhs, save_stack))):
            nm_pos = env.get(nm, pos)
            restore_shadowed(save)
            aeNegPos(rhs, nm_pos, env, res)

    elif isinstance(e, A.Tuple):
        for a in e.args:
            aeNegPos(a, pos, env, res)

    elif isinstance(e, A.LetTuple):
        # propagate negation-position flags through body
        saved = save_and_shadow(e.names)
        aeNegPos(e.body, pos, env, res)
        # merge negation position of the tuple variables
        nm_pos = [env[nm] for nm in e.names if nm in env]
        if len(nm_pos) > 0:
            pos = nm_pos[0]
            for p in nm_pos[1:]:
                if p != pos:
                    pos = "0"
        # now back-propagate to the right-hand-side
        aeNegPos(e.rhs, pos, env, res)

    else:
        assert False, "bad case"

    return res


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# SMT Solver wrapper; handles ternary logic etc.


class DebugSolverFrame:
    def __init__(self):
        self.commands = []

    def str_lines(self, show_smt=False):
        lines = []
        for c in self.commands:
            if c[0] == "bind":
                cmd, names, rhs, smt = c
                cmd = "bind    "
                for nm, r in zip(names, rhs):
                    lines.append(f"{cmd}{nm} = {r}")
                    if show_smt:
                        lines.append(f"    smt {SMT.to_smtlib(smt)}")
                    cmd = "        "
            elif c[0] == "tuplebind":
                cmd, names, rhs, smt = c
                nms = ",".join([str(n) for n in names])
                lines.append(f"bind    {nms} = {rhs}")
                assert type(smt) is tuple
                if show_smt:
                    for s in smt:
                        lines.append(f"    smt {SMT.to_smtlib(s)}")
            elif c[0] == "assume":
                cmd, e, smt = c
                lines.append(f"assume  {e}")
                if show_smt:
                    lines.append(f"    smt {SMT.to_smtlib(smt)}")
            else:
                assert False, "bad case"
        return lines

    def add_bind(self, names, rhs, smt):
        self.commands.append(("bind", names, rhs, smt))

    def add_tuple_bind(self, names, rhs, smt):
        self.commands.append(("tuplebind", names, rhs, smt))

    def add_assumption(self, e, smt_e):
        self.commands.append(("assume", e, smt_e))


@dataclass
class TernVal:
    v: Any
    d: Any

    def __str__(self):
        return f"({self.v},{self.d})"


def is_ternary(x):
    return isinstance(x, TernVal)


class SMTSolver:
    def __init__(self, verbose=False):
        self.env = ChainMap()
        self.stride_sym = ChainMap()
        self.const_sym = dict()
        self.const_sym_count = 1
        self.solver = _get_smt_solver()
        self.verbose = verbose
        self.z3 = Z3SubProc()

        self.Z3_MODE = True
        self.z3slv = z3lib.Solver()

        # used during lowering
        self.mod_div_tmp_bins = []
        self.negative_pos = None

        # debug info
        self.frames = [DebugSolverFrame()]

    def to_ternary(self, x):
        if self.Z3_MODE:
            return x if is_ternary(x) else TernVal(x, Z3.BoolVal(True))
        else:
            return x if is_ternary(x) else TernVal(x, SMT.Bool(True))

    def push(self):
        self.solver.push()
        self.z3.push()
        self.internal_push()
        self.z3slv.push()

    def pop(self):
        self.internal_pop()
        self.z3.pop()
        self.solver.pop()
        self.z3slv.pop()

    def internal_push(self):
        self.env = self.env.new_child()
        self.stride_sym = self.stride_sym.new_child()
        self.frames.append(DebugSolverFrame())

    def internal_pop(self):
        self.frames.pop()
        self.stride_sym = self.stride_sym.parents
        self.env = self.env.parents

    def debug_str(self, smt=False):
        lines = []
        for f in self.frames:
            lns = f.str_lines(show_smt=smt)
            if len(lines) > 0 and len(lns) > 0:
                lines.append("")
            lines += lns
        return "\n".join(lines)

    def _bind(self, names, rhs):
        """bind will make sure the provided names are equal to
        the provided right-hand-sides for the remainder of the
        scope that we're in"""
        for x, e in zip(names, rhs):
            smt_e = self._lower(e)
            self.env[x] = smt_e

    def _bind_tuple(self, names, rhs):
        """bind will make sure the provided names are equal to
        the provided right-hand-sides for the remainder of the
        scope that we're in.
        bind_tuple will simultaneously bind the whole list of
        names to a tuple-type right-hand-side"""
        smt_rhs = self._lower(rhs)
        for x, e in zip(names, smt_rhs):
            self.env[x] = e

    def _add_free_vars(self, e):
        fv = aeFV(e)
        # print("\n\n\nADDING FREE VAR\n")
        # print(e)
        # print(fv)
        # print()
        for x, typ in fv.items():
            if type(x) is tuple:
                x = self._get_stride_sym(*x)
            elif type(x) is ConstSymFV:
                x = self._get_const_sym(x.name)
                # print("CONST SYM", x)
            if x in self.env:
                pass  # already defined; no worries
            else:
                v = self._getvar(x, typ)  # force adding to environment
                if self.Z3_MODE:
                    pass
                else:
                    self.z3.add_var(v.symbol_name(), typ)

    def assume(self, e):
        assert e.type is T.bool
        e = e.simplify()
        self._add_free_vars(e)
        self.negative_pos = aeNegPos(e, "-")
        smt_e = self._lower(e)
        assert not is_ternary(smt_e), "assumptions must be classical"
        self.frames[-1].add_assumption(e, smt_e)
        if self.Z3_MODE:
            self.z3slv.assert_exprs(smt_e)
        else:
            self.z3.add_assertion(smt_e)
            # self.solver.add_assertion(smt_e)

    def satisfy(self, e):
        assert e.type is T.bool
        e = e.simplify()
        self.push()
        self._add_free_vars(e)
        self.negative_pos = aeNegPos(e, "-")
        smt_e = self._lower(e)
        assert not is_ternary(smt_e), "formulas must be classical"
        if self.Z3_MODE:
            self.z3slv.assert_exprs(smt_e)
            result = self.z3slv.check()
            if result == Z3.sat:
                is_sat = True
            elif result == Z3.unsat:
                is_sat = False
            else:
                raise TypeError("unknown result from z3")
        else:
            self.z3.add_assertion(smt_e)
            is_sat = self.z3.run_check_sat()
        # is_sat      = self.solver.is_sat(smt_e)
        self.pop()
        return is_sat

    def verify(self, e):
        assert e.type is T.bool
        e = e.simplify()
        self.push()
        self._add_free_vars(e)
        self.negative_pos = aeNegPos(e, "+")
        smt_e = self._lower(e)
        assert not is_ternary(smt_e), "formulas must be classical"
        if self.verbose:
            print("*******\n*******\n*******")
            print(self.debug_str(smt=False))
            print("to verify")
            print(e)
            print(smt_e)
            print("smtlib2")
            if self.Z3_MODE:
                pass  # print(Z3.to_smt2(smt_e))
            else:
                print(SMT.to_smtlib(smt_e))
        if self.Z3_MODE:
            self.z3slv.assert_exprs(Z3.Not(smt_e))
            if self.verbose and self.Z3_MODE:
                print(self.z3slv.to_smt2())
            result = self.z3slv.check()
            if result == Z3.sat:
                is_valid = False
            elif result == Z3.unsat:
                is_valid = True
            else:
                raise TypeError("unknown result from z3")
        else:
            self.z3.add_assertion(SMT.Not(smt_e))
            is_valid = not self.z3.run_check_sat()
        # is_valid    = self.solver.is_valid(smt_e)
        self.pop()
        return is_valid

    def counter_example(self):
        raise NotImplementedError("Out of Date")

        def keep_sym(s):
            if type(s) is tuple:
                return True
            else:
                return s.get_type() == SMT.INT or s.get_type() == SMT.BOOL

        env_syms = [(sym, smt) for sym, smt in self.env.items() if keep_sym(smt)]
        smt_syms = []
        for _, smt in env_syms:
            if is_ternary(smt):
                smt_syms += [smt.v, smt.d]
            else:
                smt_syms.append(smt)
        val_map = self.solver.get_py_values(smt_syms)

        mapping = dict()
        for sym, smt in env_syms:
            if is_ternary(smt):
                x, d = val_map[smt.v], val_map[smt.d]
                mapping[sym] = "unknown" if not d else x
            else:
                mapping[sym] = val_map[smt]
        return mapping

    def _get_stride_sym(self, name, dim):
        key = (name, dim)
        if key not in self.stride_sym:
            self.stride_sym[key] = Sym(f"{name}_stride_{dim}")
        sym = self.stride_sym[key]
        return sym

    def _get_const_sym(self, name):
        if name not in self.const_sym:
            self.const_sym[name] = Sym(f"CONST_{name}")  # self.const_sym_count
            # self.const_sym_count += 1
        return self.const_sym[name]
        # return self._getvar(self.const_sym[name])
        # return SMT.Int(self.const_sym[name])

    def _get_real_const(self, name):
        if name not in self.const_sym:
            self.const_sym[name] = self.const_sym_count
            self.const_sym_count += 1
        if self.Z3_MODE:
            return Z3.Int(self.const_sym[name])
        else:
            return SMT.Int(self.const_sym[name])

    def _getvar(self, sym, typ=T.index):
        if sym not in self.env:
            if self.Z3_MODE:
                if typ.is_indexable() or typ.is_stridable():
                    self.env[sym] = Z3.Int(repr(sym))
                elif typ is T.bool:
                    self.env[sym] = Z3.Bool(repr(sym))
                elif typ.is_real_scalar():
                    self.env[sym] = Z3.Int(repr(sym))
                else:
                    assert False, f"bad type: {typ}"
            else:
                if typ.is_indexable() or typ.is_stridable():
                    self.env[sym] = SMT.Symbol(repr(sym), SMT.INT)
                elif typ is T.bool:
                    self.env[sym] = SMT.Symbol(repr(sym), SMT.BOOL)
                elif typ.is_real_scalar():
                    self.env[sym] = SMT.Symbol(repr(sym), SMT.INT)
                else:
                    assert False, f"bad type: {typ}"
        return self.env[sym]

    def _newvar(self, sym, typ=T.index, ternary=False):
        """make sure that we have a new distinct copy of this name."""
        nm = repr(sym) if sym not in self.env else repr(sym.copy())
        if self.Z3_MODE:
            smt_typ = Z3.Bool if typ == T.bool else Z3.Int

            smt_sym = smt_typ(nm)
            if ternary:
                self.env[sym] = TernVal(smt_sym, Z3.Bool(nm + "_def"))
            else:
                self.env[sym] = smt_sym
            return self.env[sym]
        else:
            smt_typ = SMT.BOOL if typ == T.bool else SMT.INT

            smt_sym = SMT.Symbol(nm, smt_typ)
            if ternary:
                self.env[sym] = TernVal(smt_sym, SMT.Symbol(nm + "_def", SMT.BOOL))
            else:
                self.env[sym] = smt_sym
            return self.env[sym]

    def _add_mod_div_eq(self, new_sym, eq):
        self.mod_div_tmp_bins[-1].append((new_sym, eq))

    def _lower(self, e):
        if e.type == T.bool:
            self.mod_div_tmp_bins.append([])
        smt_e = self._lower_body(e)
        if e.type == T.bool:
            tmp_bin = self.mod_div_tmp_bins.pop()
            # possibly wrap some definitions of temporaries
            if len(tmp_bin) > 0:
                assert not is_ternary(smt_e), "TODO: handle ternary"
                all_syms = [sym for sym, eq in tmp_bin]
                if self.Z3_MODE:
                    all_eq = Z3.And(*[eq for sym, eq in tmp_bin])
                    if self.negative_pos[id(e)] == "+":
                        smt_e = Z3.ForAll(all_syms, Z3.Implies(all_eq, smt_e))
                    else:
                        smt_e = Z3.Exists(all_syms, Z3.And(all_eq, smt_e))
                else:
                    all_eq = SMT.And(*[eq for sym, eq in tmp_bin])
                    if self.negative_pos[id(e)] == "+":
                        smt_e = SMT.ForAll(all_syms, SMT.Implies(all_eq, smt_e))
                    else:
                        smt_e = SMT.Exists(all_syms, SMT.And(all_eq, smt_e))
        return smt_e

    def _lower_body(self, e):
        if isinstance(e, A.Const):
            if e.type == T.bool:
                if self.Z3_MODE:
                    return Z3.BoolVal(e.val)
                else:
                    return SMT.Bool(e.val)
            elif e.type.is_indexable():
                if self.Z3_MODE:
                    return Z3.IntVal(e.val)
                else:
                    return SMT.Int(e.val)
            elif e.type.is_real_scalar():
                return self._get_real_const(e.val)
            else:
                assert False, f"unrecognized const type: {type(e.val)}"
        elif isinstance(e, A.ConstSym):
            # convert constant symbol to a unique integer for this symbol
            assert e.type.is_real_scalar()
            return self._getvar(self._get_const_sym(e.name))
        elif isinstance(e, A.Var):
            return self._getvar(e.name, e.type)
        elif isinstance(e, A.Unk):
            if self.Z3_MODE:
                val = Z3.BoolVal(False) if e.type == T.bool else Z3.IntVal(0)
                return TernVal(val, Z3.BoolVal(False))
            else:
                val = SMT.Bool(False) if e.type == T.bool else SMT.Int(0)
                return TernVal(val, SMT.Bool(False))
        elif isinstance(e, A.Not):
            assert e.arg.type == T.bool
            a = self._lower(e.arg)
            NOT = Z3.Not if self.Z3_MODE else SMT.Not
            if is_ternary(a):
                return TernVal(NOT(a.v), a.d)
            else:
                return NOT(a)
        elif isinstance(e, A.USub):
            assert e.arg.type.is_indexable()
            a = self._lower(e.arg)
            if self.Z3_MODE:
                if is_ternary(a):
                    return TernVal(-a.v, a.d)
                else:
                    return -a
            else:
                if is_ternary(a):
                    return TernVal(SMT.Minus(SMT.Int(0), a.v), a.d)
                else:
                    return SMT.Minus(SMT.Int(0), a)
        elif isinstance(e, A.Stride):
            return self._getvar(self._get_stride_sym(e.name, e.dim))
        elif isinstance(e, A.LetStrides):
            self.internal_push()
            for i, s in enumerate(e.strides):
                stridesym = self._get_stride_sym(e.name, i)
                self.env[stridesym] = self._lower(s)
            body = self._lower(e.body)
            self.internal_pop()
            return body
        elif isinstance(e, A.Select):
            assert e.cond.type == T.bool
            cond = self._lower(e.cond)
            tcase = self._lower(e.tcase)
            fcase = self._lower(e.fcase)
            if is_ternary(cond) or is_ternary(tcase) or is_ternary(fcase):
                c = self.to_ternary(cond)
                t = self.to_ternary(tcase)
                f = self.to_ternary(fcase)
                if self.Z3_MODE:
                    return TernVal(
                        Z3.If(c.v, t.v, f.v), Z3.And(c.d, Z3.If(c.v, t.d, f.d))
                    )
                else:
                    return TernVal(
                        SMT.Ite(c.v, t.v, f.v), SMT.And(c.d, SMT.Ite(c.v, t.d, f.d))
                    )
            else:
                if self.Z3_MODE:
                    return Z3.If(cond, tcase, fcase)
                else:
                    return SMT.Ite(cond, tcase, fcase)
        elif isinstance(e, (A.ForAll, A.Exists)):
            assert e.arg.type == T.bool
            self.internal_push()
            nm = self._newvar(e.name)
            a = self._lower(e.arg)
            self.internal_pop()
            if self.Z3_MODE:
                OP = Z3.ForAll if isinstance(e, A.ForAll) else Z3.Exists
                if is_ternary(a):
                    # forall defined if (forall nm. d) \/ (exists nm. ¬a /\ d)
                    # exists defined if (forall nm. d) \/ (exists nm. a /\ d)
                    short_a = a.v if isinstance(e, A.Exists) else Z3.Not(a.v)
                    is_def = Z3.Or(
                        Z3.ForAll([nm], a.d), Z3.Exists([nm], Z3.And(short_a, a.d))
                    )
                    return TernVal(OP([nm], a.v), is_def)
                else:
                    return OP([nm], a)
            else:
                OP = SMT.ForAll if isinstance(e, A.ForAll) else SMT.Exists
                if is_ternary(a):
                    # forall defined if (forall nm. d) \/ (exists nm. ¬a /\ d)
                    # exists defined if (forall nm. d) \/ (exists nm. a /\ d)
                    short_a = a.v if isinstance(e, A.Exists) else SMT.Not(a.v)
                    is_def = SMT.Or(
                        SMT.ForAll([nm], a.d), SMT.Exists([nm], SMT.And(short_a, a.d))
                    )
                    return TernVal(OP([nm], a.v), is_def)
                else:
                    return OP([nm], a)
        elif isinstance(e, (A.Definitely, A.Maybe)):
            assert e.arg.type == T.bool
            a = self._lower(e.arg)
            if is_ternary(a):
                if self.Z3_MODE:
                    if isinstance(e, A.Definitely):
                        return Z3.And(a.v, a.d)
                    else:
                        return Z3.Or(a.v, Z3.Not(a.d))
                else:
                    if isinstance(e, A.Definitely):
                        return SMT.And(a.v, a.d)
                    else:
                        return SMT.Or(a.v, SMT.Not(a.d))
            else:
                return a
        elif isinstance(e, A.Let):
            self.internal_push()
            self._bind(e.names, e.rhs)
            body = self._lower(e.body)
            self.internal_pop()
            return body
        elif isinstance(e, A.Tuple):
            return tuple(self._lower(a) for a in e.args)
        elif isinstance(e, A.LetTuple):
            assert type(e.rhs.type) is tuple
            assert len(e.names) == len(e.rhs.type)
            self.internal_push()
            self._bind_tuple(e.names, e.rhs)
            body = self._lower(e.body)
            self.internal_pop()
            return body
        elif isinstance(e, A.BinOp):
            lhs = self._lower(e.lhs)
            rhs = self._lower(e.rhs)
            tern = is_ternary(lhs) or is_ternary(rhs)
            AND = Z3.And if self.Z3_MODE else SMT.And
            OR = Z3.Or if self.Z3_MODE else SMT.Or
            NOT = Z3.Not if self.Z3_MODE else SMT.Not
            if tern:
                lhs = self.to_ternary(lhs)
                rhs = self.to_ternary(rhs)
                lhs, dl = lhs.v, lhs.d
                rhs, dr = rhs.v, rhs.d
                dval = AND(dl, dr)  # default for int sort

            if e.op == "+":
                val = (lhs + rhs) if self.Z3_MODE else SMT.Plus(lhs, rhs)
            elif e.op == "-":
                val = (lhs - rhs) if self.Z3_MODE else SMT.Minus(lhs, rhs)
            elif e.op == "*":
                val = (lhs * rhs) if self.Z3_MODE else SMT.Times(lhs, rhs)
            elif e.op == "/":
                assert isinstance(e.rhs, A.Const)
                assert e.rhs.val > 0
                # Introduce new Sym (z in formula below)
                div_tmp = self._getvar(Sym("div_tmp"))
                # rhs*z <= lhs < rhs*(z+1)
                if self.Z3_MODE:
                    rhs_eq = rhs * div_tmp <= lhs
                    lhs_eq = lhs < rhs * (div_tmp + Z3.IntVal(1))
                else:
                    rhs_eq = SMT.LE(SMT.Times(rhs, div_tmp), lhs)
                    lhs_eq = SMT.LT(lhs, SMT.Times(rhs, SMT.Plus(div_tmp, SMT.Int(1))))
                self._add_mod_div_eq(div_tmp, AND(rhs_eq, lhs_eq))
                val = div_tmp
            elif e.op == "%":
                assert isinstance(e.rhs, A.Const)
                assert e.rhs.val > 0
                # In the below, copy the logic above for division
                # to construct `mod_tmp` s.t.
                #   mod_tmp = floor(lhs / rhs)
                # Then,
                #   lhs % rhs = lhs - rhs * mod_tmp
                mod_tmp = self._getvar(Sym("mod_tmp"))
                if self.Z3_MODE:
                    rhs_eq = rhs * mod_tmp <= lhs
                    lhs_eq = lhs < rhs * (mod_tmp + Z3.IntVal(1))
                    self._add_mod_div_eq(mod_tmp, AND(rhs_eq, lhs_eq))
                    val = lhs - rhs * mod_tmp
                else:
                    rhs_eq = SMT.LE(SMT.Times(rhs, mod_tmp), lhs)
                    lhs_eq = SMT.LT(lhs, SMT.Times(rhs, SMT.Plus(mod_tmp, SMT.Int(1))))
                    self._add_mod_div_eq(mod_tmp, AND(rhs_eq, lhs_eq))
                    val = SMT.Minus(lhs, SMT.Times(rhs, mod_tmp))

            elif e.op == "<":
                val = (lhs < rhs) if self.Z3_MODE else SMT.LT(lhs, rhs)
            elif e.op == ">":
                val = (lhs > rhs) if self.Z3_MODE else SMT.GT(lhs, rhs)
            elif e.op == "<=":
                val = (lhs <= rhs) if self.Z3_MODE else SMT.LE(lhs, rhs)
            elif e.op == ">=":
                val = (lhs >= rhs) if self.Z3_MODE else SMT.GE(lhs, rhs)
            elif e.op == "==":
                if e.lhs.type == T.bool and e.rhs.type == T.bool:
                    val = (lhs == rhs) if self.Z3_MODE else SMT.Iff(lhs, rhs)
                elif e.lhs.type.is_indexable() and e.rhs.type.is_indexable():
                    val = (lhs == rhs) if self.Z3_MODE else SMT.Equals(lhs, rhs)
                elif e.lhs.type.is_stridable() and e.rhs.type.is_stridable():
                    val = (lhs == rhs) if self.Z3_MODE else SMT.Equals(lhs, rhs)
                elif e.lhs.type == e.rhs.type:
                    assert e.lhs.type.is_real_scalar()
                    val = (lhs == rhs) if self.Z3_MODE else SMT.Equals(lhs, rhs)
                else:
                    assert False, "bad case"
            elif e.op == "and":
                val = AND(lhs, rhs)
                if tern:
                    dval = OR(AND(dl, dr), AND(NOT(lhs), dl), AND(NOT(rhs), dr))
            elif e.op == "or":
                val = OR(lhs, rhs)
                if tern:
                    dval = OR(AND(dl, dr), AND(lhs, dl), AND(rhs, dr))
            elif e.op == "==>":
                IMPLIES = Z3.Implies if self.Z3_MODE else SMT.Implies
                val = IMPLIES(lhs, rhs)
                if tern:
                    dval = OR(AND(dl, dr), AND(NOT(lhs), dl), AND(rhs, dr))

            else:
                assert False, f"bad op: {e.op}"

            return TernVal(val, dval) if tern else val

        else:
            assert False, f"bad case: {type(e)}"


class Z3SubProc:
    def __init__(self):
        self.stack_lines = [[]]  # list(i.e. stack) of lists
        pass

    def push(self):
        self.stack_lines.append([])

    def pop(self):
        self.stack_lines.pop()

    def _get_whole_str(self):
        return "\n".join([line for frame in self.stack_lines for line in frame])

    def add_var(self, varname, vartyp):
        sort = "Bool" if vartyp == T.bool else "Int"
        self.stack_lines[-1].append(f"(declare-const {varname} {sort})")

    def add_assertion(self, smt_formula):
        smtstr = SMT.to_smtlib(smt_formula)
        self.stack_lines[-1].append(f"(assert {smtstr})")

    def run_check_sat(self):
        slv = z3lib.Solver()
        slv.from_string(self._get_whole_str())
        result = slv.check()
        if result == z3lib.z3.sat:
            return True
        elif result == z3lib.z3.unsat:
            return False
        else:
            raise Error("unknown result from z3")


# install simplify
from . import analysis_simplify



from . import query_asts as QAST
from .LoopIR import LoopIR, T
from .prelude import *


@extclass(QAST.Call)
def __str__(self):
    return f"{self.proc}(_)"


del __str__


@extclass(QAST.Alloc)
def __str__(self):
    return f"{self.name} : _"


del __str__


@extclass(QAST.WriteConfig)
def __str__(self):
    return f"{self.config.name()}.{self.field} = {str(self.rhs)}"


del __str__


@extclass(QAST.Assign)
@extclass(QAST.Reduce)
def __str__(self):
    if len(self.idx) > 0:
        return f"{self.name}[_] = {str(self.rhs)}"
    else:
        return f"{self.name} = {str(self.rhs)}"


del __str__


@extclass(QAST.For)
def __str__(self):
    return f"for {self.name} in par(0, {str(self.hi)}):_"


del __str__


@extclass(QAST.If)
def __str__(self):
    cond = str(self.cond)
    return f"if {cond}:_"


del __str__


@extclass(QAST.Read)
def __str__(self):
    if len(self.idx) > 0:
        return f"{self.name}[_]"
    else:
        return f"{self.name}"


del __str__


@extclass(QAST.Const)
def __str__(self):
    return f"{self.val}"


del __str__


@extclass(QAST.USub)
def __str__(self):
    return f"-{str(self.arg)}"


del __str__


@extclass(QAST.BinOp)
def __str__(self):
    lhs = str(self.lhs)
    rhs = str(self.rhs)
    return f"{lhs} {self.op} {rhs}"


del __str__


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Conversion from LoopIR AST to QAST


class LoopIR_to_QAST:
    def __init__(self, loopir_node):
        self.loopir_node = loopir_node

        self.names = dict()

        if isinstance(loopir_node, LoopIR.proc):
            self.qast = self.map_proc(loopir_node)
        elif isinstance(loopir_node, list):
            if len(loopir_node) == 0:
                self.qast = []
            elif isinstance(loopir_node[0], LoopIR.stmt):
                self.qast = self.map_stmts(loopir_node)
            else:
                assert False, f"cannot process list of {type(loopir_node[0])}"
        elif isinstance(loopir_node, LoopIR.fnarg):
            self.qast = self.map_fnarg(loopir_node)
        elif isinstance(loopir_node, LoopIR.stmt):
            self.qast = self.map_stmt(loopir_node)
        elif isinstance(loopir_node, LoopIR.expr):
            self.qast = self.map_expr(loopir_node)
        elif isinstance(loopir_node, LoopIR.type):
            self.qast = self.map_type(loopir_node)

    def result(self):
        return self.qast

    def getname(self, sym):
        return str(sym)

    def bindname(self, sym):
        return str(sym)

    def map_proc(self, proc):
        name = proc.name
        return QAST.Proc(
            name,
            [self.map_fnarg(fa) for fa in proc.args],
            [self.map_expr(p) for p in proc.preds],
            self.map_stmts(proc.body),
            proc.instr,
        )

    def map_fnarg(self, fa):
        name = self.bindname(fa.name)
        return QAST.FnArg(name, self.map_type(fa.type), fa.mem)

    def map_stmts(self, body):
        return [self.map_stmt(s) for s in body]

    def map_stmt(self, s):
        styp = type(s)
        if styp is LoopIR.Assign or styp is LoopIR.Reduce:
            qtyp = QAST.Assign if styp is LoopIR.Assign else QAST.Reduce
            name = self.getname(s.name)
            return qtyp(
                name,
                self.map_type(s.type),
                [self.map_expr(i) for i in s.idx],
                self.map_expr(s.rhs),
            )
        elif styp is LoopIR.WriteConfig:
            return QAST.WriteConfig(s.config, s.field, self.map_expr(s.rhs))
        elif styp is LoopIR.Pass:
            return QAST.Pass()
        elif styp is LoopIR.If:
            return QAST.If(
                self.map_expr(s.cond), self.map_stmts(s.body), self.map_stmts(s.orelse)
            )
        elif styp is LoopIR.For:
            name = self.bindname(s.iter)
            return QAST.For(
                name,
                self.map_expr(s.lo),
                self.map_expr(s.hi),
                self.map_stmts(s.body),
                False,
            )
        elif styp is LoopIR.Alloc:
            name = self.bindname(s.name)
            return QAST.Alloc(name, self.map_type(s.type), s.mem)
        elif styp is LoopIR.Call:
            return QAST.Call(s.f.name, [self.map_expr(a) for a in s.args])
        elif styp is LoopIR.WindowStmt:
            name = self.bindname(s.name)
            return QAST.WindowStmt(name, self.map_expr(s.rhs))
        else:
            assert False, f"bad case: {styp}"

    def map_expr(self, e):
        etyp = type(e)
        if etyp is LoopIR.Read:
            return QAST.Read(
                self.getname(e.name),
                [self.map_expr(i) for i in e.idx],
                self.map_type(e.type),
            )
        elif etyp is LoopIR.Const:
            return QAST.Const(e.val, self.map_type(e.type))
        elif etyp is LoopIR.USub:
            return QAST.USub(self.map_expr(e.arg), self.map_type(e.type))
        elif etyp is LoopIR.BinOp:
            return QAST.BinOp(
                e.op, self.map_expr(e.lhs), self.map_expr(e.rhs), self.map_type(e.type)
            )
        elif etyp is LoopIR.BuiltIn:
            return QAST.BuiltIn(
                e.f.name(), [self.map_expr(a) for a in e.args], self.map_type(e.type)
            )
        elif etyp is LoopIR.WindowExpr:
            name = self.getname(e.name)

            def map_w(w):
                if isinstance(w, LoopIR.Interval):
                    return QAST.Interval(self.map_expr(w.lo), self.map_expr(w.hi))
                else:
                    return QAST.Point(self.map_expr(w.pt))

            return QAST.WindowExpr(
                name, [map_w(w) for w in e.idx], self.map_type(e.type)
            )
        elif etyp is LoopIR.StrideExpr:
            name = self.getname(e.name)
            return QAST.StrideExpr(name, e.dim, self.map_type(e.type))
        elif etyp is LoopIR.ReadConfig:
            return QAST.ReadConfig(e.config, e.field, self.map_type(e.type))
        else:
            assert False, f"bad case: {etyp}"

    def map_type(self, typ):
        if typ == T.R:
            return QAST.R()
        elif typ == T.f16:
            return QAST.f16()
        elif typ == T.f32:
            return QAST.f32()
        elif typ == T.f64:
            return QAST.f64()
        elif typ == T.i8:
            return QAST.i8()
        elif typ == T.i32:
            return QAST.i32()
        elif typ == T.bool:
            return QAST.bool()
        elif typ == T.int:
            return QAST.int()
        elif typ == T.index:
            return QAST.index()
        elif typ == T.size:
            return QAST.size()
        elif typ == T.stride:
            return QAST.stride()
        elif typ.is_tensor_or_window():
            as_tensor = typ.as_tensor if isinstance(typ, T.Window) else typ
            return QAST.tensor(
                [self.map_expr(e) for e in as_tensor.hi],
                as_tensor.is_window,
                self.map_type(as_tensor.type),
            )
        else:
            assert False, f"bad case: {type(typ)}"



from __future__ import annotations

from exo.stdlib.scheduling import *
from exo import proc
from exo.range_analysis import (
    constant_bound,
    arg_range_analysis,
    IndexRangeEnvironment,
)
from exo.stdlib.range_analysis import (
    infer_range,
    bounds_inference,
)
from exo.LoopIR import LoopIR, T


def test_affine_index_range():
    @proc
    def bar():
        for i in seq(0, 6):
            for j in seq(0, i + 2):
                pass

    e = bar.find("for j in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    e_range = constant_bound(e, {i_sym: (0, 5)})
    assert e_range == (2, 7)


def test_affine_index_range1():
    @proc
    def bar():
        for i in seq(0, 6):
            for j in seq(0, (i + 2) * 5):
                pass

    e = bar.find("for j in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    e_range = constant_bound(e, {i_sym: (0, 5)})
    assert e_range == (10, 35)


def test_affine_index_range2():
    @proc
    def bar(N: size):
        assert N >= 1
        assert N <= 5
        for i in seq(0, 6):
            for j in seq(0, N + (i + 2) * 5):
                pass

    e = bar.find("for j in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    N_sym = bar._loopir_proc.args[0].name
    e_range = constant_bound(e, {i_sym: (0, 5), N_sym: (1, 5)})
    assert e_range == (11, 40)


def test_affine_index_range3():
    @proc
    def bar(N: size):
        assert N >= 1
        assert N <= 5
        for i in seq(0, 6):
            for j in seq(0, (N + (i + 2) * 5) / 2):
                pass

    e = bar.find("for j in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    N_sym = bar._loopir_proc.args[0].name
    e_range = constant_bound(e, {i_sym: (0, 5), N_sym: (1, 5)})
    assert e_range == (5, 20)


def test_affine_index_range4():
    @proc
    def bar(N: size):
        assert N >= 1
        assert N <= 5
        for i in seq(0, 6):
            for j in seq(0, (N + (i + 2) * 5) - 4):
                pass

    e = bar.find("for j in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    N_sym = bar._loopir_proc.args[0].name
    e_range = constant_bound(e, {i_sym: (0, 5), N_sym: (1, 5)})
    assert e_range == (7, 36)


def test_affine_index_range5():
    @proc
    def bar(N: size):
        assert N >= 1
        assert N <= 5
        for i in seq(0, 6):
            for j in seq(0, (i + 2) * 5 - N):
                pass

    e = bar.find("for j in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    N_sym = bar._loopir_proc.args[0].name
    e_range = constant_bound(e, {i_sym: (0, 5), N_sym: (1, 5)})
    assert e_range == (5, 34)


def test_affine_index_range6():
    @proc
    def bar(N: size):
        assert N >= 1
        assert N <= 5
        for i in seq(0, 6):
            for j in seq(0, (i + 2) * 5 - N):
                pass

    e = bar.find("for j in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    N_sym = bar._loopir_proc.args[0].name
    e_range = constant_bound(e, {i_sym: (0, 5), N_sym: (1, 5)})
    assert e_range == (5, 34)


def test_affine_index_range7():
    @proc
    def bar(N: size):
        assert N >= 1
        assert N <= 5
        for i in seq(0, 6):
            for j in seq(0, (-3) + (i + 2) * 5 - N):
                pass

    e = bar.find("for j in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    N_sym = bar._loopir_proc.args[0].name
    e_range = constant_bound(e, {i_sym: (0, 5), N_sym: (1, 5)})
    assert e_range == (2, 31)


def test_affine_index_range8():
    @proc
    def bar():
        for i in seq(0, 6):
            for j in seq(0, (i + 10) % 4):
                pass

    e = bar.find("for j in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    e_range = constant_bound(e, {i_sym: (0, 5)})
    assert e_range == (0, 3)


def test_affine_index_range9():
    @proc
    def bar():
        for i in seq(0, 6):
            for j in seq(0, (i + 10) % 9):
                pass

    e = bar.find("for j in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    e_range = constant_bound(e, {i_sym: (0, 5)})
    assert e_range == (1, 6)


def test_affine_index_range10():
    @proc
    def bar():
        for i in seq(0, 6):
            for j in seq(0, 8):
                for k in seq(0, 2 * i + 3 * j):
                    pass

    e = bar.find("for k in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    j_sym = bar.find("for j in _:_")._impl._node.iter
    e_range = constant_bound(e, {i_sym: (0, 5), j_sym: (0, 7)})
    assert e_range == (0, 31)


def test_affine_index_range_fail():
    @proc
    def bar():
        for i in seq(0, 6):
            for j in seq(0, (-i) * 3 + i * 4):
                pass

    e = bar.find("for j in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    e_range = constant_bound(e, {i_sym: (0, 5)})
    assert e_range == (-15, 20)


def test_affine_index_range_fail2():
    @proc
    def bar(N: size):
        for i in seq(0, 6):
            for j in seq(0, (i - 2) / 2 + 10):
                pass

    e = bar.find("for j in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    e_range = constant_bound(e, {i_sym: (0, 5)})
    assert e_range == (9, 11)


def test_affine_index_range_fail3():
    @proc
    def bar():
        for i in seq(0, 3):
            for j in seq(0, i * 16 + 16 - i * 16):
                pass

    e = bar.find("for j in _:_").hi()._impl._node
    i_sym = bar.find("for i in _:_")._impl._node.iter
    e_range = constant_bound(e, {i_sym: (0, 2)})
    assert e_range == (-16, 48)


def test_arg_range():
    @proc
    def foo(N: size):
        pass

    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[0], fast=False
    ) == (1, None)
    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[0], fast=True
    ) == (1, None)


def test_arg_range2():
    @proc
    def foo(N: size, K: size):
        assert N >= 50
        assert K > 20
        pass

    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[0], fast=False
    ) == (50, None)
    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[1], fast=False
    ) == (21, None)
    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[0], fast=True
    ) == (1, None)


def test_arg_range3():
    @proc
    def foo(N: size, K: size):
        assert N == 50
        assert K > 20
        assert K >= 100
        pass

    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[0], fast=False
    ) == (50, 50)
    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[1], fast=False
    ) == (100, None)


def test_arg_range4():
    @proc
    def foo(N: size, K: size):
        assert N < 500
        assert K > 20
        assert K >= 100000
        pass

    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[0], fast=False
    ) == (1, 499)
    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[1], fast=False
    ) == (
        2**15,
        None,
    )


def test_arg_range5():
    @proc
    def foo(N: size, K: size):
        assert N < 500
        assert K > 20
        assert K >= 100000
        pass

    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[0], fast=False
    ) == (1, 499)
    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[1], fast=False
    ) == (
        2**15,
        None,
    )


def test_arg_range6():
    val = 2**32

    @proc
    def foo(N: size, K: size):
        assert N > val
        assert K < val
        pass

    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[0], fast=False
    ) == (
        2**15,
        None,
    )
    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[1], fast=False
    ) == (1, None)


def test_arg_range7():
    @proc
    def foo(N: index, K: index):
        assert -10 < K
        assert K < 4
        pass

    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[0], fast=False
    ) == (
        None,
        None,
    )
    assert arg_range_analysis(
        foo._loopir_proc, foo._loopir_proc.args[1], fast=False
    ) == (-9, 3)


def test_index_range_env():
    N_upper_bound = 20
    K_lower_bound = 30
    M_value = 100

    @proc
    def foo(N: size, K: size, M: size):
        assert N <= N_upper_bound
        assert K >= K_lower_bound
        assert M == M_value
        for i in seq(K, N * 2 + K):
            pass
        for j in seq(M, M + 1):
            pass

    env = IndexRangeEnvironment(foo._loopir_proc, fast=False)
    loop = foo.find_loop("i")
    node = loop._impl._node

    N_read = LoopIR.Read(
        foo._loopir_proc.args[0].name, [], T.size, foo._loopir_proc.srcinfo
    )

    def run_N_asserts():
        assert env.check_expr_bounds(
            0,
            IndexRangeEnvironment.leq,
            N_read,
            IndexRangeEnvironment.leq,
            N_upper_bound,
        )
        assert env.check_expr_bounds(
            1,
            IndexRangeEnvironment.leq,
            N_read,
            IndexRangeEnvironment.lt,
            N_upper_bound + 1,
        )
        assert not env.check_expr_bounds(
            1,
            IndexRangeEnvironment.lt,
            N_read,
            IndexRangeEnvironment.lt,
            N_upper_bound + 1,
        )
        assert not env.check_expr_bounds(
            1, IndexRangeEnvironment.lt, N_read, IndexRangeEnvironment.lt, N_upper_bound
        )
        assert not env.check_expr_bounds(
            N_upper_bound,
            IndexRangeEnvironment.eq,
            N_read,
            IndexRangeEnvironment.eq,
            N_upper_bound,
        )

    run_N_asserts()

    env.enter_scope()

    # I should still be able to see `N`
    run_N_asserts()

    env.add_loop_iter(node.iter, node.lo, node.hi)
    i_read = LoopIR.Read(node.iter, [], T.index, foo._loopir_proc.srcinfo)

    run_N_asserts()

    assert not env.check_expr_bounds(
        0, IndexRangeEnvironment.leq, i_read, IndexRangeEnvironment.leq, 100
    )
    assert not env.check_expr_bounds(
        K_lower_bound,
        IndexRangeEnvironment.leq,
        i_read,
        IndexRangeEnvironment.leq,
        100000,
    )
    assert env.check_expr_bound(0, IndexRangeEnvironment.leq, i_read)
    assert env.check_expr_bound(K_lower_bound, IndexRangeEnvironment.leq, i_read)
    assert not env.check_expr_bound(K_lower_bound, IndexRangeEnvironment.lt, i_read)

    env.exit_scope()

    run_N_asserts()

    loop = foo.find_loop("j")
    node = loop._impl._node
    env.add_loop_iter(node.iter, node.lo, node.hi)
    N_read = LoopIR.Read(
        foo._loopir_proc.args[0].name, [], T.size, foo._loopir_proc.srcinfo
    )
    j_read = LoopIR.Read(node.iter, [], T.index, foo._loopir_proc.srcinfo)
    assert env.check_expr_bounds(
        M_value, IndexRangeEnvironment.eq, j_read, IndexRangeEnvironment.eq, M_value
    )
    assert env.check_expr_bounds(
        M_value, IndexRangeEnvironment.eq, j_read, IndexRangeEnvironment.leq, M_value
    )
    assert not env.check_expr_bounds(
        M_value, IndexRangeEnvironment.eq, j_read, IndexRangeEnvironment.lt, M_value
    )


# TODO: write test for IndexRange


def test_infer_range_of_expr():
    # TODO: test loops with non-zero lower bounds
    @proc
    def foo(n: size, m: size, x: i8[n + 2, m + 5]):
        assert n % 4 == 0
        assert m % 3 == 0
        for io in seq(0, n / 4):
            for jo in seq(0, m / 3):
                for ii in seq(0, 6):
                    for ji in seq(0, 8):
                        x[io * 4 + ii, jo * 3 + ji] = 1.0

    idx_c = foo.find("x[_] = _").idx()[0]
    idx = idx_c._impl._node

    bounds = infer_range(idx_c, foo.find_loop("jo"))
    assert str(bounds) == "(io * 4, 0, 5)"

    bounds = infer_range(idx_c, foo.find_loop("io"))
    assert str(bounds) == "(io * 4, 0, 5)"


def test_bounds_inference_tiled_blur2d():
    @proc
    def blur2d_tiled(n: size, consumer: i8[n, n], sin: i8[n + 1, n + 1]):
        assert n % 4 == 0
        producer: i8[n + 1, n + 1]
        for i in seq(0, n + 1):
            for j in seq(0, n + 1):
                producer[i, j] = sin[i, j]
        for io in seq(0, n / 4):
            for jo in seq(0, n / 4):
                for ii in seq(0, 4):
                    for ji in seq(0, 4):
                        consumer[4 * io + ii, 4 * jo + ji] = (
                            producer[4 * io + ii, 4 * jo + ji]
                            + producer[4 * io + ii, 4 * jo + ji + 1]
                            + producer[4 * io + ii + 1, 4 * jo + ji]
                            + producer[4 * io + ii + 1, 4 * jo + ji + 1]
                        ) / 4.0

    loop = blur2d_tiled.find_loop("io")
    io_sym = loop._impl._node.iter
    bound = bounds_inference(blur2d_tiled, loop, "producer", 0, include=["R"])
    assert str(bound) == "(io * 4, 0, 4)"

    loop = blur2d_tiled.find_loop("jo")
    jo_sym = loop._impl._node.iter
    bound = bounds_inference(blur2d_tiled, loop, "producer", 1, include=["R"])
    assert str(bound) == "(jo * 4, 0, 4)"

    bound = bounds_inference(blur2d_tiled, loop, "producer", 0, include=["R"])
    assert str(bound) == "(io * 4, 0, 4)"
    assert bound.get_stride_of(io_sym) == 4
    bound = bounds_inference(blur2d_tiled, loop, "producer", 1, include=["R"])
    assert str(bound) == "(jo * 4, 0, 4)"
    assert bound.get_stride_of(jo_sym) == 4


def test_bounds_inference_fail():
    @proc
    def foo(n: size, x: i8[n]):
        for i in seq(0, n):
            for j in seq(0, n):
                x[i] = 1.0
                x[j] = 1.0

    loop = foo.find_loop("j")
    bound = bounds_inference(foo, loop, "x", 0, include=["W"])
    assert str(bound) == "(0, -inf, inf)"



from __future__ import annotations

import dataclasses
import enum
import functools
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import Optional


class InvalidCursorError(Exception):
    pass


def _starts_with(a: list, b: list):
    """
    Returns true if the first elements of `a` equal `b` exactly
    >>> _starts_with([1, 2, 3], [1, 2])
    True
    >>> _starts_with(['x'], ['x'])
    True
    >>> _starts_with([1, 2, 3], [])
    True
    >>> _starts_with(['a', 'b', 'c'], ['a', 'b', 'c', 'd'])
    False
    """
    return len(a) >= len(b) and all(x == y for x, y in zip(a, b))


def _is_sub_range(a: range, b: range):
    """
    Returns true if `a` is a STRICT sub-range of `b`.
    Only applies to step-1 ranges.
    >>> _is_sub_range(range(1, 4), range(1, 4))
    False
    >>> _is_sub_range(range(0, 3), range(3, 6))
    False
    >>> _is_sub_range(range(2, 4), range(2, 5))
    True
    >>> _is_sub_range(range(2, 4), range(1, 4))
    True
    >>> _is_sub_range(range(2, 4), range(1, 5))
    True
    >>> _is_sub_range(range(0, 4), range(1, 4))
    False
    """
    assert a.step == b.step and a.step in (1, None)
    return (a.start >= b.start) and (a.stop <= b.stop) and a != b


def _intersects_partially(a: range, b: range):
    assert a.step == b.step and a.step in (1, None)
    return a.start < b.start < a.stop < b.stop or b.start < a.start < b.stop < a.stop


def forward_identity(p, fwd=None):
    fwd = fwd or (lambda x: x)

    @functools.wraps(fwd)
    def forward(cursor):
        cursor = fwd(cursor)
        if isinstance(cursor, Gap):
            return dataclasses.replace(
                cursor, _root=p, _anchor=dataclasses.replace(cursor._anchor, _root=p)
            )
        elif isinstance(cursor, Node):
            return dataclasses.replace(cursor, _root=p)
        else:
            raise InvalidCursorError("cannot forward blocks")

    return forward


@dataclass
class Cursor(ABC):
    _root: object

    # ------------------------------------------------------------------------ #
    # Static constructors
    # ------------------------------------------------------------------------ #

    @staticmethod
    def create(obj: object):
        return Node(obj, [])

    # ------------------------------------------------------------------------ #
    # Validating accessors
    # ------------------------------------------------------------------------ #

    def get_root(self):
        return self.root()._node

    # ------------------------------------------------------------------------ #
    # Navigation (universal)
    # ------------------------------------------------------------------------ #

    def root(self) -> Node:
        """Get a cursor to the root of the tree this cursor resides in"""
        return Node(self._root, [])

    # ------------------------------------------------------------------------ #
    # Navigation (abstract)
    # ------------------------------------------------------------------------ #

    @abstractmethod
    def parent(self) -> Node:
        """Get the node containing the current cursor"""

    # ------------------------------------------------------------------------ #
    # Protected path / mutation helpers
    # ------------------------------------------------------------------------ #

    def _local_forward(self, new_root, fwd_node, fwd_block):
        """
        Creates a forwarding function for "local" edits to the AST.
        Here, local means that all affected nodes share a common LCA,
        taking into account details about block identity (i.e. body
        vs orelse). It is assumed that the edited locale is the block
        in which the present cursor resides.

        The fwd_node function is applied to one of this cursor's siblings
        and is expected to return a list of new tree edges which are adjusted
        for the edit. See the implementation of insert and delete for some
        simple examples, or wrap for a more complex example (that lengthens
        certain paths).

        The fwd_block function is applied to a range of this cursor's children
        and is expected to return a list of new tree edges which are adjusted
        for the edit. This differs from fwd_node in that the last element of the
        list is of the form (attr, range) instead of (attr, index).
        """
        orig_root = self._root

        edit_path = self.parent()._path

        if isinstance(self, Node):
            attr = self._path[-1][0]
        elif isinstance(self, Gap):
            attr = self.anchor()._path[-1][0]
        else:
            assert isinstance(self, Block)
            attr = self._attr

        depth = len(edit_path)

        def forward(cursor: Cursor) -> Cursor:
            if cursor._root is not orig_root:
                raise InvalidCursorError("cannot forward from unknown root")

            if isinstance(cursor, Gap):
                return Gap(new_root, forward(cursor.anchor()), cursor.type())

            assert isinstance(cursor, (Node, Block))

            def evolve(c, **kwargs):
                return dataclasses.replace(c, _root=new_root, **kwargs)

            if isinstance(cursor, Block):
                if cursor._anchor._path == edit_path and cursor._attr == attr:
                    # Block is directly in edit scope
                    blk_path = fwd_block(attr, cursor._range)
                    new_attr, new_rng = blk_path[-1]

                    return evolve(
                        cursor,
                        _anchor=evolve(
                            cursor._anchor, _path=cursor._anchor._path + blk_path[:-1]
                        ),
                        _attr=new_attr,
                        _range=new_rng,
                    )

                # Otherwise, just forward the anchor
                try:
                    return evolve(cursor, _anchor=forward(cursor._anchor))
                except InvalidCursorError:
                    raise InvalidCursorError("block no longer exists (parent deleted)")

            old_path = cursor._path

            if len(old_path) < depth + 1:
                # Too shallow
                return evolve(cursor)

            old_attr, old_idx = old_path[depth]

            if not (_starts_with(old_path, edit_path) and old_attr == attr):
                # Different path down tree
                return evolve(cursor)

            new_path = (
                old_path[:depth] + fwd_node(attr, old_idx) + old_path[depth + 1 :]
            )
            return evolve(cursor, _path=new_path)

        return forward


@dataclass
class Block(Cursor):
    _anchor: Node
    _attr: str  # must be 'body' or 'orelse'
    _range: range

    # ------------------------------------------------------------------------ #
    # Navigation (implementation)
    # ------------------------------------------------------------------------ #

    def parent(self) -> Node:
        return self._anchor

    def depth(self) -> int:
        return self._anchor.depth()

    def before(self) -> Gap:
        return self[0].before()

    def after(self) -> Gap:
        return self[-1].after()

    # ------------------------------------------------------------------------ #
    # Container interface implementation
    # ------------------------------------------------------------------------ #

    def __contains__(self, cur):
        if isinstance(cur, Block):
            return (
                cur._anchor == self._anchor
                and cur._attr == self._attr
                and (
                    _is_sub_range(cur._range, self._range) or cur._range == self._range
                )
            )
        elif isinstance(cur, Gap):
            # This shouldn't recurse forever...
            return not cur.is_edge() and cur.anchor() in self
        elif isinstance(cur, Node):
            return (
                cur.parent() == self._anchor
                and cur._path[-1][0] == self._attr
                and cur._path[-1][1] in self._range
            )
        else:
            raise TypeError(f"cannot check containment of {type(cur)}")

    # ------------------------------------------------------------------------ #
    # Sequence interface implementation
    # ------------------------------------------------------------------------ #

    def __iter__(self):
        block = self.parent()
        for i in self._range:
            yield block._child_node(self._attr, i)

    def __getitem__(self, i):
        r = self._range[i]
        if isinstance(r, range):
            if r.step != 1:
                raise IndexError("block cursors must be contiguous")
            return Block(self._root, self._anchor, self._attr, r)
        else:
            return self._anchor._child_node(self._attr, r)

    def __len__(self):
        return len(self._range)

    # ------------------------------------------------------------------------ #
    # Block-specific operations
    # ------------------------------------------------------------------------ #

    def expand(self, delta_lo=None, delta_hi=None):
        """
        [delta_lo] or [delta_hi] being [None] means expand as much as possible to the left
        or right, respectively.
        """
        full_block = self.parent()._child_block(self._attr)
        full_range = full_block._range

        delta_lo = self._range.start if delta_lo is None else delta_lo
        delta_hi = len(full_range) - self._range.stop if delta_hi is None else delta_hi

        lo = max(0, self._range.start - delta_lo)
        hi = min(len(full_range), self._range.stop + delta_hi)

        return Block(self._root, self._anchor, self._attr, range(lo, hi))

    # ------------------------------------------------------------------------ #
    # AST mutation
    # ------------------------------------------------------------------------ #

    def _replace(self, nodes: list, *, empty_default=None):
        """
        This is an UNSAFE internal function for replacing a block in an AST with
        a list of statements and providing a forwarding function as collateral.
        It is meant to be package-private, not class-private, so it may be
        called from other internal classes and modules, but not from end-user
        code.
        """
        assert isinstance(nodes, list)

        def update(n):
            r = self._range
            children = getattr(n, self._attr)
            new_children = children[: r.start] + nodes + children[r.stop :]
            new_children = new_children or empty_default or []
            return n.update(**{self._attr: new_children})

        p = self._anchor._rewrite(update)
        fwd = self._forward_replace(p, len(nodes))
        return p, fwd

    def _forward_replace(self, new_proc, n_ins):
        del_range = self._range
        n_diff = n_ins - len(del_range)

        idx_update = lambda i: i + n_diff * (i >= del_range.stop)

        def fwd_node(attr, i):
            if i in del_range:
                raise InvalidCursorError("node no longer exists")
            return [(attr, idx_update(i))]

        def fwd_block(attr, rng):
            if _intersects_partially(rng, del_range) or _is_sub_range(rng, del_range):
                raise InvalidCursorError("block no longer exists")

            return [(attr, range(idx_update(rng.start), idx_update(rng.stop)))]

        return self._local_forward(new_proc, fwd_node, fwd_block)

    def _delete(self):
        """
        This is an UNSAFE internal function for deleting a block in an AST and
        providing a forwarding function as collateral. It is meant to be
        package-private, not class-private, so it may be called from other
        internal classes and modules, but not from end-user code.
        """
        # TODO: refactor this; LoopIR should not be imported here
        from exo.LoopIR import LoopIR

        pass_stmt = [LoopIR.Pass(None, self.parent()._node.srcinfo)]
        return self._replace([], empty_default=pass_stmt)

    def resolve_all(self):
        """
        Do this rather than `[n._node for n in self]` because that would
        walk down the tree once per node in the block, whereas this walks
        down once.
        """
        return getattr(self._anchor._node, self._attr)[
            self._range.start : self._range.stop
        ]

    def _wrap(self, ctor, wrap_attr):
        """
        This is an UNSAFE internal function for wrapping a block in an AST
        with another block-containing node and providing a forwarding function
        as collateral. It is meant to be package-private, not class-private,
        so it may be called from other internal classes and modules, but not
        from end-user code.
        """
        nodes = self.resolve_all()
        new_node = ctor(**{wrap_attr: nodes})

        def update(parent):
            r = self._range
            children = getattr(parent, self._attr)
            new_children = children[: r.start] + [new_node] + children[r.stop :]
            return parent.update(**{self._attr: new_children})

        p = self._anchor._rewrite(update)
        fwd = self._forward_wrap(p, wrap_attr)
        return p, fwd

    def _forward_wrap(self, p, wrap_attr):
        rng = self._range
        n_delta = len(rng) - 1

        def fwd_node(attr, i):
            if i >= rng.stop:
                return [(attr, i - n_delta)]
            elif i >= rng.start:
                return [(attr, rng.start), (wrap_attr, i - rng.start)]
            else:
                return [(attr, i)]

        def fwd_block(attr, blk_rng):
            if blk_rng.start >= rng.stop:
                new_rng = range(blk_rng.start - n_delta, blk_rng.stop - n_delta)
                return [(attr, new_rng)]
            elif blk_rng.stop <= rng.start:
                return [(attr, blk_rng)]
            elif blk_rng.start in rng and blk_rng.stop - 1 in rng:
                new_rng = range(blk_rng.start - rng.start, blk_rng.stop - rng.start)
                return [
                    (attr, blk_rng.start),
                    (wrap_attr, new_rng),
                ]
            elif rng.start in blk_rng and rng.stop - 1 in blk_rng:
                return [(attr, range(blk_rng.start, blk_rng.stop - len(rng) + 1))]

            # We could arguably try to forward to something?
            raise InvalidCursorError("block no longer exists")

        return self._local_forward(p, fwd_node, fwd_block)

    def _move(self, target: Gap):
        """
        This is an UNSAFE internal function for relocating a block in an AST
        and providing a forwarding function as collateral. It is meant to be
        package-private, not class-private, so it may be called from other
        internal classes and modules, but not from end-user code.
        """

        if target in self:
            target = self.before()

        nodes = self.resolve_all()

        def _is_before(g: Gap, b: Block):
            b_path = b._anchor._path + [(b._attr, b._range.start)]

            g_path = g._anchor._path[:]
            g_path[-1] = (g_path[-1][0], g._insertion_index())

            for (g_attr, g_idx), (b_attr, b_idx) in zip(g_path, b_path):
                if g_attr != b_attr:
                    # arbitrary because they're in disjoint branches
                    return False

                if g_idx != b_idx:
                    return g_idx < b_idx

            return True

        def reroot(x):
            return dataclasses.replace(x, _root=ir)

        # The following implementation "unsafely" coerces a cursor along the
        # intermediate procs by ordering the edits so that identity-forwarding
        # is actually safe. This is somewhat simpler to reason about than a
        # recursive function that has to walk down two branches simultaneously.
        # Intuition: do the "later" edit first, then the "earlier" one.
        if _is_before(target, self):
            # If the gap comes first in a pre-order traversal, then we want to
            # delete the original block of nodes first, to keep the path to the
            # gap stable, before inserting the nodes in the new position.
            ir, _ = self._delete()
            ir, _ = Gap(ir, reroot(target._anchor), target._type)._insert(nodes)
        else:
            # Conversely, if the moved block comes first, then we want to jump
            # ahead and insert the block into the gap position before coming back
            # to delete the original nodes, so that the path to the deletion stays
            # stable.
            ir, _ = target._insert(nodes)
            ir, _ = Block(ir, reroot(self._anchor), self._attr, self._range)._delete()

        fwd = self._forward_move(ir, target)
        return ir, fwd

    def _forward_move(self, p, target: Gap):
        orig_root = self._root

        block_path = self._anchor._path
        block_n = len(block_path)

        block_attr = self._attr

        blk_rng = self._range
        edit_n = len(blk_rng)

        gap_path = target._anchor._path[:]
        gap_path[-1] = (gap_path[-1][0], target._insertion_index())
        gap_n = len(gap_path) - 1

        def forward(cursor: Node):
            # This is duplicated in _local_forward. If there ever is a third
            # place where this is needed, it should be refactored into a
            # helper function.
            if cursor._root is not orig_root:
                raise InvalidCursorError("cannot forward from unknown root")

            if isinstance(cursor, Gap):
                return Gap(p, forward(cursor.anchor()), cursor.type())

            if isinstance(cursor, Block):
                anchor = cursor._anchor
                attr = cursor._attr
                rng = cursor._range
                if anchor._path == block_path and attr == block_attr:
                    if _intersects_partially(rng, blk_rng):
                        raise InvalidCursorError(
                            "move cannot forward block because exactly one endpoint intersects with moved block"
                        )

                new_start_node = forward(anchor._child_node(attr, rng.start))
                new_stop_node = forward(anchor._child_node(attr, rng.stop - 1))

                assert new_start_node.parent() == new_stop_node.parent()
                new_anchor = new_start_node.parent()

                attr1, new_start = new_start_node._path[-1]
                attr2, new_end = new_stop_node._path[-1]
                assert attr1 == attr2
                assert new_start <= new_end

                return dataclasses.replace(
                    cursor,
                    _root=p,
                    _anchor=new_anchor,
                    _range=range(new_start, new_end + 1),
                )

            assert isinstance(cursor, Node)

            cur_path = list(cursor._path)
            cur_n = len(cur_path)

            # Handle nodes around the edit points
            offsets = []

            if (
                cur_n > block_n
                and block_path == cur_path[:block_n]
                and block_attr == cur_path[block_n][0]
            ):
                if blk_rng.stop <= cur_path[block_n][1]:
                    # if after orig. block end, subtract edit_n
                    offsets.append((block_n, -edit_n))
                elif blk_rng.start <= cur_path[block_n][1]:
                    new_gap_path = gap_path
                    if block_n <= gap_n:
                        # compute new gap_path
                        block_start_path = block_path + [(block_attr, blk_rng.start)]

                        new_gap_path = []
                        for lca_n, (bs, gs) in enumerate(
                            zip(block_start_path, gap_path)
                        ):
                            if bs != gs:
                                if bs[0] == gs[0] and bs[1] < gs[1]:
                                    new_gap_path.append((gs[0], gs[1] - edit_n))
                                else:
                                    new_gap_path.append(gs)
                                break

                            new_gap_path.append(gs)

                        new_gap_path.extend(gap_path[lca_n + 1 :])

                    # if inside orig block, move to gap location
                    off = cur_path[block_n][1] - blk_rng.start
                    return dataclasses.replace(
                        cursor,
                        _root=p,
                        _path=(
                            new_gap_path[:-1]
                            + [(new_gap_path[-1][0], new_gap_path[-1][1] + off)]
                            + cur_path[block_n + 1 :]
                        ),
                    )
                else:
                    # before orig block, do nothing
                    pass

            # if after orig. gap, add edit_n
            if (
                cur_n > gap_n
                and gap_path[:gap_n] == cur_path[:gap_n]
                and gap_path[gap_n][0] == cur_path[gap_n][0]
                and gap_path[gap_n][1] <= cur_path[gap_n][1]
            ):
                offsets.append((gap_n, edit_n))

            for off_i, off_d in offsets:
                cur_path[off_i] = (cur_path[off_i][0], cur_path[off_i][1] + off_d)

            return dataclasses.replace(cursor, _root=p, _path=cur_path)

        return forward


@dataclass
class Node(Cursor):
    _path: list[tuple[str, Optional[int]]]

    # ------------------------------------------------------------------------ #
    # Validating accessors
    # ------------------------------------------------------------------------ #

    @cached_property
    def _node(self):
        """
        Gets the raw underlying node that's pointed-to. This is meant to be
        compiler-internal, not class-private, so other parts of the compiler
        may call this, while users should not.
        """
        n = self._root

        for attr, idx in self._path:
            n = getattr(n, attr)
            if idx is not None:
                n = n[idx]

        return n

    # ------------------------------------------------------------------------ #
    # Navigation (implementation)
    # ------------------------------------------------------------------------ #

    def parent(self) -> Node:
        if not self._path:
            raise InvalidCursorError("cursor does not have a parent")
        return Node(self._root, self._path[:-1])

    def depth(self) -> int:
        return len(self._path)

    def before(self) -> Gap:
        return Gap(self._root, self, GapType.Before)

    def after(self) -> Gap:
        return Gap(self._root, self, GapType.After)

    def prev(self, dist=1) -> Node:
        return self.next(-dist)

    def next(self, dist=1) -> Node:
        if not self._path:
            raise InvalidCursorError("cannot move root cursor")
        attr, i = self._path[-1]
        if i is None:
            raise InvalidCursorError("cursor is not inside block")
        return self.parent()._child_node(attr, i + dist)

    # ------------------------------------------------------------------------ #
    # Navigation (children)
    # ------------------------------------------------------------------------ #

    def _child_node(self, attr, i=None) -> Node:
        _node = getattr(self._node, attr)
        if i is not None:
            if 0 <= i < len(_node):
                _node = _node[i]
            else:
                raise InvalidCursorError("cursor is out of range")
        elif isinstance(_node, list):
            raise ValueError("must index into block attribute")
        cur = Node(self._root, self._path + [(attr, i)])
        # noinspection PyPropertyAccess
        # cached_property is settable, bug in static analysis
        cur._node = _node
        return cur

    def _child_block(self, attr: str):
        stmts = getattr(self._node, attr)
        assert isinstance(stmts, list)
        return Block(self._root, self, attr, range(len(stmts)))

    # ------------------------------------------------------------------------ #
    # Navigation (block selectors)
    # ------------------------------------------------------------------------ #

    def body(self) -> Block:
        return self._child_block("body")

    def orelse(self) -> Block:
        return self._child_block("orelse")

    # ------------------------------------------------------------------------ #
    # Conversions
    # ------------------------------------------------------------------------ #

    def as_block(self) -> Block:
        attr, i = self._path[-1]
        if i is None:
            raise InvalidCursorError("node is not inside a block")
        return Block(self._root, self.parent(), attr, range(i, i + 1))

    # ------------------------------------------------------------------------ #
    # Location queries
    # ------------------------------------------------------------------------ #

    def get_index(self):
        _, i = self._path[-1]
        return i

    def is_ancestor_of(self, other: Cursor) -> bool:
        """Return true if this node is an ancestor of another"""
        if not isinstance(other, Node):
            other = other._anchor
        return _starts_with(other._path, self._path)

    # ------------------------------------------------------------------------ #
    # AST mutation
    # ------------------------------------------------------------------------ #

    def _rewrite(self, fn):
        """
        Applies `fn` to the current node and rewrites the tree with the result.
        """

        def impl(node, path, j=0):
            if j == len(path):
                return fn(node)

            attr, i = path[j]
            children = getattr(node, attr)

            if i is None:
                return node.update(**{attr: impl(children, path, j + 1)})

            new_nodes = impl(children[i], path, j + 1)
            if not isinstance(new_nodes, list):
                new_nodes = [new_nodes]
            return node.update(**{attr: children[:i] + new_nodes + children[i + 1 :]})

        return impl(self.get_root(), self._path)

    def _delete(self):
        return self.as_block()._delete()

    def _move(self, gap: Gap):
        return self.as_block()._move(gap)

    def _replace(self, ast):
        """
        This is an UNSAFE internal function for replacing a node in an AST with
        either a list of statements or another single node and providing a
        forwarding function as collateral. It is meant to be package-private,
        not class-private, so it may be called from other internal classes and
        modules, but not from end-user code.
        """
        attr, idx = self._path[-1]
        if idx is not None and isinstance(ast, list):
            # noinspection PyProtectedMember
            # delegate block replacement to the Block class
            return self.as_block()._replace(ast)

        # replacing a single expression, or something not in a block
        assert not isinstance(ast, list), "replaced node is not in a block"

        p = self._rewrite(lambda _: ast)
        fwd = self._forward_replace(p)
        return p, fwd

    def _forward_replace(self, new_root, can_fwd_node=True):
        def fwd_node(*_):
            return self._path[-1:]

        def fwd_block(attr, rng):
            return [(attr, rng)]

        return self._local_forward(new_root, fwd_node, fwd_block)


class GapType(enum.Enum):
    Before = 0
    After = 1
    # BodyStart = enum.auto()
    # BodyEnd = enum.auto()
    # OrElseStart = enum.auto()
    # OrElseEnd = enum.auto()


@dataclass
class Gap(Cursor):
    _anchor: Node
    _type: GapType

    # ------------------------------------------------------------------------ #
    # Navigation (implementation)
    # ------------------------------------------------------------------------ #

    def parent(self) -> Node:
        if self.is_edge():
            return self._anchor
        return self._anchor.parent()

    def depth(self) -> int:
        return self._anchor.depth()

    def anchor(self) -> Node:
        return self._anchor

    # ------------------------------------------------------------------------ #
    # Gap type queries
    # ------------------------------------------------------------------------ #

    def is_edge(self):
        return self._type not in (GapType.Before, GapType.After)

    def type(self) -> GapType:
        return self._type

    # ------------------------------------------------------------------------ #
    # AST mutation
    # ------------------------------------------------------------------------ #

    def _insert(self, stmts: list):
        """
        This is an UNSAFE internal function for inserting a list of nodes at a
        particular gap in an AST block and providing a forwarding function as
        collateral. It is meant to be package-private, not class-private, so it
        may be called from other internal classes and modules, but not from
        end-user code.
        """

        def update(anchor):
            if self._type == GapType.Before:
                return stmts + [anchor]
            elif self._type == GapType.After:
                return [anchor] + stmts
            else:
                assert False, f"case {self._type} not implemented"

        if self.is_edge():
            raise NotImplementedError()

        p = self.anchor()._rewrite(update)
        fwd = self._forward_insert(p, len(stmts))
        return p, fwd

    def _forward_insert(self, new_root, ins_len):
        ins_idx = self._insertion_index()

        idx_update = lambda i: i + ins_len * (i >= ins_idx)

        def fwd_node(attr, i):
            return [(attr, idx_update(i))]

        def fwd_block(attr, rng):
            new_rng = range(
                idx_update(rng.start),
                idx_update(rng.stop - 1) + 1,
            )
            return [(attr, new_rng)]

        return self._local_forward(new_root, fwd_node, fwd_block)

    def _insertion_index(self):
        _, i = self._anchor._path[-1]
        if self._type == GapType.Before:
            return i
        elif self._type == GapType.After:
            return i + 1
        else:
            assert False, f"case {self.type} not implemented"



from __future__ import annotations

import ast as pyast
import inspect
import re
import sys
import textwrap
from collections import ChainMap

from asdl_adt.validators import ValidationError

from .API_types import ProcedureBase
from .builtins import *
from .configs import Config
from .LoopIR import UAST, PAST, front_ops
from .prelude import *


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Helpers


class ParseError(Exception):
    pass


class SizeStub:
    def __init__(self, nm):
        assert isinstance(nm, Sym)
        self.nm = nm


def str_to_mem(name):
    return getattr(sys.modules[__name__], name)


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Top-level decorator


def get_ast_from_python(f):
    # note that we must dedent in case the function is defined
    # inside of a local scope

    rawsrc = inspect.getsource(f)
    src = textwrap.dedent(rawsrc)
    n_dedent = len(re.match("^(.*)", rawsrc).group()) - len(
        re.match("^(.*)", src).group()
    )
    srcfilename = inspect.getsourcefile(f)
    _, srclineno = inspect.getsourcelines(f)
    srclineno -= 1  # adjust for decorator line

    # create way to query for src-code information
    def getsrcinfo(node):
        return SrcInfo(
            filename=srcfilename,
            lineno=node.lineno + srclineno,
            col_offset=node.col_offset + n_dedent,
            end_lineno=(
                None if node.end_lineno is None else node.end_lineno + srclineno
            ),
            end_col_offset=(
                None if node.end_col_offset is None else node.end_col_offset + n_dedent
            ),
        )

    # convert into AST nodes; which should be a module with a single node
    module = pyast.parse(src)
    assert len(module.body) == 1

    return module.body[0], getsrcinfo


def get_src_locals(*, depth):
    """
    Get global and local environments for context capture purposes
    """
    stack_frames: [inspect.FrameInfo] = inspect.stack()
    assert len(stack_frames) >= depth
    func_locals = stack_frames[depth].frame.f_locals
    assert isinstance(func_locals, dict)
    return ChainMap(func_locals)


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Pattern-Parser top-level, invoked on strings rather than as a decorator


def pattern(s, filename=None, lineno=None):
    assert isinstance(s, str)

    src = s
    n_dedent = 0
    if filename is not None:
        srcfilename = filename
        srclineno = lineno
    else:
        srcfilename = "string"
        srclineno = 0

    module = pyast.parse(src)
    assert isinstance(module, pyast.Module)

    # create way to query for src-code information
    def getsrcinfo(node):
        return SrcInfo(
            filename=srcfilename,
            lineno=node.lineno + srclineno,
            col_offset=node.col_offset + n_dedent,
            end_lineno=(
                None if node.end_lineno is None else node.end_lineno + srclineno
            ),
            end_col_offset=(
                None if node.end_col_offset is None else node.end_col_offset + n_dedent
            ),
        )

    parser = Parser(module.body, getsrcinfo, is_fragment=True)
    return parser.result()


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Parser Pass object


class Parser:
    def __init__(
        self,
        module_ast,
        getsrcinfo,
        is_fragment=False,
        func_globals=None,
        srclocals=None,
        as_func=False,
        as_config=False,
        instr=None,
    ):

        self.module_ast = module_ast
        self.globals = func_globals
        self.locals = srclocals or ChainMap()
        self.getsrcinfo = getsrcinfo
        self.is_fragment = is_fragment

        self.push()

        builtins = {"sin": sin, "relu": relu, "select": select}
        if is_fragment:
            self.AST = PAST
        else:
            self.AST = UAST
            # add builtins
            for key, val in builtins.items():
                self.locals[key] = val

        if as_func:
            self._cached_result = self.parse_fdef(module_ast, instr=instr)
        elif as_config:
            self._cached_result = self.parse_cls(module_ast)
        elif is_fragment:
            is_expr = False
            if len(module_ast) == 1:
                s = module_ast[0]
                special_cases = list(builtins.keys()) + ["stride"]
                if isinstance(s, pyast.Expr) and (
                    not isinstance(s.value, pyast.Call)
                    or s.value.func.id in special_cases
                ):
                    is_expr = True

            if is_expr:
                self._cached_result = self.parse_expr(s.value)
            else:
                self._cached_result = self.parse_stmt_block(module_ast)
        else:
            assert False, "parser mode configuration unsupported"
        self.pop()

    def result(self):
        return self._cached_result

    def push(self):
        self.locals = self.locals.new_child()

    def pop(self):
        self.locals = self.locals.parents

    # - # - # - # - # - # - # - # - # - # - # - # - # - # - # - #
    # parser helper routines

    def err(self, node, errstr, origin=None):
        raise ParseError(f"{self.getsrcinfo(node)}: {errstr}") from origin

    def eval_expr(self, expr):
        assert isinstance(expr, pyast.expr)
        code = compile(pyast.Expression(expr), "", "eval")
        e_obj = eval(code, self.globals, self.locals)
        return e_obj

    # - # - # - # - # - # - # - # - # - # - # - # - # - # - # - #
    # structural parsing rules...

    def parse_fdef(self, fdef, instr=None):
        assert isinstance(fdef, pyast.FunctionDef)

        fargs = fdef.args
        bad_arg_syntax_errmsg = """
    Exo expects function arguments to not use these python features:
      - position-only arguments
      - unnamed (position or keyword) arguments (i.e. *varargs, **kwargs)
      - keyword-only arguments
      - default argument values
    """
        if (
            len(fargs.posonlyargs) > 0
            or fargs.vararg is not None
            or len(fargs.kwonlyargs) > 0
            or len(fargs.kw_defaults) > 0
            or fargs.kwarg is not None
            or len(fargs.defaults) > 0
        ):
            self.err(fargs, bad_arg_syntax_errmsg)

        # process each argument in order
        # we will assume for now that all sizes come first
        # sizes = []
        args = []
        names = set()
        for a in fargs.args:
            if a.annotation is None:
                self.err(a, "expected argument to be typed, i.e. 'x : T'")
            tnode = a.annotation

            typ, mem = self.parse_arg_type(tnode)
            if a.arg in names:
                self.err(a, f"repeated argument name: '{a.arg}'")
            names.add(a.arg)
            nm = Sym(a.arg)
            if isinstance(typ, UAST.Size):
                self.locals[a.arg] = SizeStub(nm)
            else:
                # note we don't need to stub the index variables
                self.locals[a.arg] = nm
            args.append(UAST.fnarg(nm, typ, mem, self.getsrcinfo(a)))

        # return types are non-sensical for Exo, b/c it models procedures
        if fdef.returns is not None:
            self.err(fdef, "Exo does not support function return types")

        # parse out any assertions at the front of the statement block
        # first split the assertions out
        first_non_assert = 0
        while first_non_assert < len(fdef.body) and isinstance(
            fdef.body[first_non_assert], pyast.Assert
        ):
            first_non_assert += 1
        assertions = fdef.body[0:first_non_assert]
        pyast_body = fdef.body[first_non_assert:]

        # then parse out predicates from the assertions
        preds = []
        for a in assertions:
            if a.msg is not None:
                self.err(a, "Exo procedure assertions should not have messages")

            # stride-expr handling
            a = a.test
            preds.append(self.parse_expr(a))

        if instr:
            instr = UAST.instr(*instr)
        # parse the procedure body
        body = self.parse_stmt_block(pyast_body)
        return UAST.proc(
            name=fdef.name,
            args=args,
            preds=preds,
            body=body,
            instr=instr,
            srcinfo=self.getsrcinfo(fdef),
        )

    def parse_cls(self, cls):
        assert isinstance(cls, pyast.ClassDef)

        name = cls.name
        if len(cls.bases) > 0:
            self.err(cls, f"expected no base classes in a config definition")
        # ignore cls.keywords and cls.decorator_list

        fields = [self.parse_config_field(stmt) for stmt in cls.body]

        return (name, fields)

    def parse_config_field(self, stmt):
        basic_err = "expected config field definition of the form: " "name : type"
        if isinstance(stmt, pyast.AnnAssign):
            if stmt.value:
                self.err(stmt, basic_err)
            elif not isinstance(stmt.target, pyast.Name):
                self.err(stmt, basic_err)
            name = stmt.target.id

            typ_list = {
                "bool": UAST.Bool(),
                "size": UAST.Size(),
                "index": UAST.Index(),
                "stride": UAST.Stride(),
            }
            for k in Parser._prim_types:
                typ_list[k] = Parser._prim_types[k]
            del typ_list["R"]

            if (
                isinstance(stmt.annotation, pyast.Name)
                and stmt.annotation.id in typ_list
            ):
                typ = typ_list[stmt.annotation.id]
            else:
                self.err(
                    stmt.annotation,
                    "expected one of the following "
                    "types: " + ", ".join(list(typ_list.keys())),
                )

            return name, typ
        else:
            self.err(stmt, basic_err)

    def parse_arg_type(self, node):
        # Arg was of the form ` name : annotation `
        # The annotation will be of one of the following forms
        #   ` type `
        #   ` type @ memory `
        def is_at(x):
            return isinstance(x, pyast.BinOp) and isinstance(x.op, pyast.MatMult)

        # decompose top-level of annotation syntax
        if is_at(node):
            typ_node = node.left
            mem_node = node.right
        else:
            typ_node = node
            mem_node = None

        # detect which sort of type we have here
        is_size = lambda x: isinstance(x, pyast.Name) and x.id == "size"
        is_index = lambda x: isinstance(x, pyast.Name) and x.id == "index"
        is_bool = lambda x: isinstance(x, pyast.Name) and x.id == "bool"
        is_stride = lambda x: isinstance(x, pyast.Name) and x.id == "stride"

        # parse each kind of type here
        if is_size(typ_node):
            if mem_node is not None:
                self.err(
                    node, "size types should not be annotated with " "memory locations"
                )
            return UAST.Size(), None

        elif is_index(typ_node):
            if mem_node is not None:
                self.err(
                    node, "size types should not be annotated with " "memory locations"
                )
            return UAST.Index(), None

        elif is_bool(typ_node):
            if mem_node is not None:
                self.err(
                    node, "size types should not be annotated with " "memory locations"
                )
            return UAST.Bool(), None

        elif is_stride(typ_node):
            if mem_node is not None:
                self.err(
                    node,
                    "stride types should not be annotated with " "memory locations",
                )
            return UAST.Stride(), None

        else:
            typ = self.parse_num_type(typ_node, is_arg=True)

            mem = self.eval_expr(mem_node) if mem_node else None

            return typ, mem

    def parse_alloc_typmem(self, node):
        if isinstance(node, pyast.BinOp) and isinstance(node.op, pyast.MatMult):
            # node.right == Name
            # x[n] @ DRAM
            # x[n] @ lib.scratch
            mem = self.eval_expr(node.right)
            node = node.left
        else:
            mem = None
        typ = self.parse_num_type(node)
        return typ, mem

    _prim_types = {
        "R": UAST.Num(),
        "f16": UAST.F16(),
        "f32": UAST.F32(),
        "f64": UAST.F64(),
        "i8": UAST.INT8(),
        "ui8": UAST.UINT8(),
        "ui16": UAST.UINT16(),
        "i32": UAST.INT32(),
    }

    def parse_num_type(self, node, is_arg=False):
        if isinstance(node, pyast.Subscript):
            if isinstance(node.value, pyast.List):
                if is_arg is not True:
                    self.err(
                        node,
                        "Window expression such as [R] "
                        "should only be used in the function "
                        "signature",
                    )
                if len(node.value.elts) != 1:
                    self.err(
                        node,
                        "Window expression should annotate " "only one type, e.g. [R]",
                    )

                base = node.value.elts[0]
                if (
                    not isinstance(base, pyast.Name)
                    or base.id not in Parser._prim_types
                ):
                    self.err(
                        node,
                        "expected window type to be of "
                        "the form '[R][...]', '[f32][...]', etc.",
                    )

                typ = Parser._prim_types[base.id]
                is_window = True
            elif (
                isinstance(node.value, pyast.Name)
                and node.value.id in Parser._prim_types
            ):
                typ = Parser._prim_types[node.value.id]
                is_window = False
            else:
                self.err(
                    node,
                    "expected tensor type to be "
                    "of the form 'R[...]', 'f32[...]', etc.",
                )

            if sys.version_info[:3] >= (3, 9):
                # unpack single or multi-arg indexing to list of slices/indices
                if isinstance(node.slice, pyast.Slice):
                    self.err(node, "index-slicing not allowed")
                else:
                    if isinstance(node.slice, pyast.Tuple):
                        dims = node.slice.elts
                    else:
                        dims = [node.slice]
            else:
                if isinstance(node.slice, (pyast.Slice, pyast.ExtSlice)):
                    self.err(node, "index-slicing not allowed")
                else:
                    assert isinstance(node.slice, pyast.Index)
                    if isinstance(node.slice.value, pyast.Tuple):
                        dims = node.slice.value.elts
                    else:
                        dims = [node.slice.value]

            # convert the dimension list into a full tensor type
            exprs = [self.parse_expr(idx) for idx in dims]
            typ = UAST.Tensor(exprs, is_window, typ)

            return typ

        elif isinstance(node, pyast.Name) and node.id in Parser._prim_types:
            return Parser._prim_types[node.id]
        else:
            self.err(node, "unrecognized type: " + ast.dump(node))

    def parse_stmt_block(self, stmts):
        assert isinstance(stmts, list)

        rstmts = []

        for s in stmts:
            # ----- Assginment, Reduction, Var Declaration/Allocation parsing
            if isinstance(s, (pyast.Assign, pyast.AnnAssign, pyast.AugAssign)):
                # parse the rhs first, if it's present
                rhs = None
                if isinstance(s, pyast.AnnAssign):
                    if s.value is not None:
                        self.err(
                            s, "Variable declaration should not " "have value assigned"
                        )
                    if self.is_fragment:
                        name_node, idxs, _ = self.parse_lvalue(s.target)
                        if len(idxs) > 0:
                            self.err(s.target, "expected simple name in declaration")
                        nm = name_node.id
                        if nm != "_" and not is_valid_name(nm):
                            self.err(name_node, "expected valid name or _")
                        _, sizes, _ = self.parse_array_indexing(s.annotation)
                        rstmts.append(PAST.Alloc(nm, sizes, self.getsrcinfo(s)))
                        continue  # escape rest of case
                else:
                    rhs = self.parse_expr(s.value)

                # parse the lvalue expression
                if isinstance(s, pyast.Assign):
                    if len(s.targets) > 1:
                        self.err(
                            s,
                            "expected only one expression "
                            "on the left of an assignment",
                        )
                    node = s.targets[0]
                    # handle WriteConfigs
                    if isinstance(node, pyast.Attribute):
                        if not isinstance(node.value, pyast.Name):
                            self.err(
                                node,
                                "expected configuration writes "
                                "of the form 'config.field = ...'",
                            )
                        assert isinstance(node.attr, str)

                        if self.is_fragment:
                            assert isinstance(node.value.id, str)
                            rstmts.append(
                                PAST.WriteConfig(
                                    node.value.id, node.attr, self.getsrcinfo(s)
                                )
                            )
                        else:
                            # lookup config and early-exit
                            config_obj = self.eval_expr(node.value)
                            if not isinstance(config_obj, Config):
                                self.err(
                                    node.value,
                                    "expected indexed object " "to be a Config",
                                )

                            # early-exit in this case
                            field_name = node.attr
                            rstmts.append(
                                UAST.WriteConfig(
                                    config_obj, field_name, rhs, self.getsrcinfo(s)
                                )
                            )
                        continue
                    # handle all other lvalue cases
                    else:
                        lvalue_tmp = self.parse_lvalue(node)
                        name_node, idxs, is_window = lvalue_tmp
                    lhs = s.targets[0]
                else:
                    name_node, idxs, is_window = self.parse_lvalue(s.target)
                    lhs = s.target

                if self.is_fragment:
                    # check that the name is valid
                    nm = name_node.id
                    if nm != "_" and not is_valid_name(nm):
                        self.err(name_node, "expected valid name or _")

                    # generate the assignemnt or reduction statement
                    if isinstance(s, pyast.Assign):
                        rstmts.append(PAST.Assign(nm, idxs, rhs, self.getsrcinfo(s)))
                    elif isinstance(s, pyast.AugAssign):
                        if not isinstance(s.op, pyast.Add):
                            self.err(s, "only += reductions currently supported")
                        rstmts.append(PAST.Reduce(nm, idxs, rhs, self.getsrcinfo(s)))
                else:
                    if is_window:
                        self.err(
                            lhs,
                            "cannot perform windowing on "
                            "left-hand-side of an assignment",
                        )
                    if isinstance(s, pyast.AnnAssign) and len(idxs) > 0:
                        self.err(lhs, "expected simple name in declaration")

                    # insert any needed Allocs
                    if isinstance(s, pyast.AnnAssign):
                        nm = Sym(name_node.id)
                        self.locals[name_node.id] = nm
                        typ, mem = self.parse_alloc_typmem(s.annotation)
                        rstmts.append(UAST.Alloc(nm, typ, mem, self.getsrcinfo(s)))

                    # handle cases of ambiguous assignment to undefined
                    # variables
                    if (
                        isinstance(s, pyast.Assign)
                        and len(idxs) == 0
                        and name_node.id not in self.locals
                    ):
                        nm = Sym(name_node.id)
                        self.locals[name_node.id] = nm
                        do_fresh_assignment = True
                    else:
                        do_fresh_assignment = False

                    # get the symbol corresponding to the name on the
                    # left-hand-side
                    if isinstance(s, (pyast.Assign, pyast.AugAssign)):
                        if name_node.id not in self.locals:
                            self.err(name_node, f"variable '{name_node.id}' undefined")
                        nm = self.locals[name_node.id]
                        if isinstance(nm, SizeStub):
                            self.err(
                                name_node,
                                f"cannot write to size variable '{name_node.id}'",
                            )
                        elif not isinstance(nm, Sym):
                            self.err(
                                name_node,
                                f"expected '{name_node.id}' to "
                                f"refer to a local variable",
                            )

                    # generate the assignemnt or reduction statement
                    if do_fresh_assignment:
                        rstmts.append(UAST.FreshAssign(nm, rhs, self.getsrcinfo(s)))
                    elif isinstance(s, pyast.Assign):
                        rstmts.append(UAST.Assign(nm, idxs, rhs, self.getsrcinfo(s)))
                    elif isinstance(s, pyast.AugAssign):
                        if not isinstance(s.op, pyast.Add):
                            self.err(s, "only += reductions currently supported")
                        rstmts.append(UAST.Reduce(nm, idxs, rhs, self.getsrcinfo(s)))

            # ----- For Loop parsing
            elif isinstance(s, pyast.For):
                if len(s.orelse) > 0:
                    self.err(s, "else clause on for-loops unsupported")

                self.push()

                if not isinstance(s.target, pyast.Name):
                    self.err(s.target, "expected simple name for iterator variable")

                if self.is_fragment:
                    itr = s.target.id
                else:
                    itr = Sym(s.target.id)
                    self.locals[s.target.id] = itr

                cond = self.parse_loop_cond(s.iter)
                body = self.parse_stmt_block(s.body)

                if self.is_fragment:
                    lo, hi = cond
                    rstmts.append(PAST.For(itr, lo, hi, body, self.getsrcinfo(s)))
                else:
                    rstmts.append(UAST.For(itr, cond, body, self.getsrcinfo(s)))

                self.pop()

            # ----- If statement parsing
            elif isinstance(s, pyast.If):
                cond = self.parse_expr(s.test)

                self.push()
                body = self.parse_stmt_block(s.body)
                self.pop()
                self.push()
                orelse = self.parse_stmt_block(s.orelse)
                self.pop()

                rstmts.append(self.AST.If(cond, body, orelse, self.getsrcinfo(s)))

            # ----- Sub-routine call parsing
            elif (
                isinstance(s, pyast.Expr)
                and isinstance(s.value, pyast.Call)
                and isinstance(s.value.func, pyast.Name)
            ):
                if self.is_fragment:
                    # handle stride expression
                    if s.value.func.id == "stride":
                        if (
                            len(s.value.keywords) > 0
                            or len(s.value.args) != 2
                            or not isinstance(s.value.args[0], pyast.Name)
                            or not isinstance(s.value.args[1], pyast.Constant)
                            or not isinstance(s.value.args[1].value, int)
                        ):
                            self.err(
                                s.value,
                                "expected stride(...) to "
                                "have exactly 2 arguments: the identifier "
                                "for the buffer we are talking about "
                                "and an integer specifying which dimension",
                            )

                        name = s.value.args[0].id
                        dim = int(s.value.args[1].value)

                        rstmts.append(
                            PAST.StrideExpr(name, dim, self.getsrcinfo(s.value))
                        )
                    else:
                        if len(s.value.keywords) > 0:
                            self.err(
                                s.value,
                                "cannot call procedure() " "with keyword arguments",
                            )

                        args = [self.parse_expr(a) for a in s.value.args]

                        rstmts.append(
                            PAST.Call(s.value.func.id, args, self.getsrcinfo(s.value))
                        )
                else:
                    f = self.eval_expr(s.value.func)
                    if not isinstance(f, ProcedureBase):
                        self.err(
                            s.value.func, f"expected called object " "to be a procedure"
                        )

                    if len(s.value.keywords) > 0:
                        self.err(
                            s.value, "cannot call procedure() " "with keyword arguments"
                        )

                    args = [self.parse_expr(a) for a in s.value.args]

                    rstmts.append(
                        UAST.Call(f.INTERNAL_proc(), args, self.getsrcinfo(s.value))
                    )

            # ----- Pass no-op parsing
            elif isinstance(s, pyast.Pass):
                rstmts.append(self.AST.Pass(self.getsrcinfo(s)))

            # ----- Stmt Hole parsing
            elif (
                isinstance(s, pyast.Expr)
                and isinstance(s.value, pyast.Name)
                and s.value.id == "_"
            ):
                rstmts.append(PAST.S_Hole(self.getsrcinfo(s.value)))

            elif isinstance(s, pyast.Assert):
                self.err(
                    s,
                    "predicate assert should happen at the beginning " "of a function",
                )
            else:
                self.err(s, "unsupported type of statement")

        return rstmts

    def parse_loop_cond(self, cond):
        if isinstance(cond, pyast.Call):
            if isinstance(cond.func, pyast.Name) and cond.func.id in ("par", "seq"):
                if len(cond.keywords) > 0:
                    self.err(
                        cond, "par() and seq() does not support" " named arguments"
                    )
                elif len(cond.args) != 2:
                    self.err(cond, "par() and seq() expects exactly" " 2 arguments")
                lo = self.parse_expr(cond.args[0])
                hi = self.parse_expr(cond.args[1])

                if self.is_fragment:
                    return lo, hi
                else:
                    if cond.func.id == "par":
                        return UAST.ParRange(lo, hi, self.getsrcinfo(cond))
                    else:
                        return UAST.SeqRange(lo, hi, self.getsrcinfo(cond))
            else:
                self.err(
                    cond,
                    "expected for loop condition to be in the form "
                    "'par(...,...)' or 'seq(...,...)'",
                )
        else:
            e_hole = PAST.E_Hole(self.getsrcinfo(cond))
            if self.is_fragment:
                return e_hole, e_hole
            return e_hole

    # parse the left-hand-side of an assignment
    def parse_lvalue(self, node):
        if not isinstance(node, (pyast.Name, pyast.Subscript)):
            self.err(node, "expected lhs of form 'x' or 'x[...]'")
        else:
            return self.parse_array_indexing(node)

    def parse_array_indexing(self, node):
        if isinstance(node, pyast.Name):
            return node, [], False
        elif isinstance(node, pyast.Subscript):
            if sys.version_info[:3] >= (3, 9):
                # unpack single or multi-arg indexing to list of slices/indices
                if isinstance(node.slice, pyast.Tuple):
                    dims = node.slice.elts
                else:
                    dims = [node.slice]
            else:
                if isinstance(node.slice, pyast.Slice):
                    dims = [node.slice]
                elif isinstance(node.slice, pyast.ExtSlice):
                    dims = node.slice.dims
                else:
                    assert isinstance(node.slice, pyast.Index)
                    if isinstance(node.slice.value, pyast.Tuple):
                        dims = node.slice.value.elts
                    else:
                        dims = [node.slice.value]

            if not isinstance(node.value, pyast.Name):
                self.err(node, "expected access to have form 'x' or 'x[...]'")

            is_window = any(isinstance(e, pyast.Slice) for e in dims)
            idxs = [
                (self.parse_slice(e, node) if is_window else self.parse_expr(e))
                for e in dims
            ]

            return node.value, idxs, is_window
        else:
            assert False, "bad case"

    def parse_slice(self, e, node):
        assert not self.is_fragment, "Window PAST unsupported"

        if sys.version_info[:3] >= (3, 9):
            srcinfo = self.getsrcinfo(e)
        else:
            if isinstance(e, pyast.Index):
                e = e.value
                srcinfo = self.getsrcinfo(e)
            else:
                srcinfo = self.getsrcinfo(node)

        if isinstance(e, pyast.Slice):
            lo = None if e.lower is None else self.parse_expr(e.lower)
            hi = None if e.upper is None else self.parse_expr(e.upper)
            if e.step is not None:
                self.err(
                    e,
                    "expected windowing to have the form x[:], "
                    "x[i:], x[:j], or x[i:j], but not x[i:j:k]",
                )

            return UAST.Interval(lo, hi, srcinfo)
        else:
            return UAST.Point(self.parse_expr(e), srcinfo)

    # parse expressions, including values, indices, and booleans
    def parse_expr(self, e):
        if isinstance(e, (pyast.Name, pyast.Subscript)):
            nm_node, idxs, is_window = self.parse_array_indexing(e)

            if self.is_fragment:
                nm = nm_node.id
                if len(idxs) == 0 and nm == "_":
                    return PAST.E_Hole(self.getsrcinfo(e))
                else:
                    return PAST.Read(nm, idxs, self.getsrcinfo(e))
            else:
                if nm_node.id in self.locals:
                    nm = self.locals[nm_node.id]
                elif nm_node.id in self.globals:
                    nm = self.globals[nm_node.id]
                else:
                    nm = None

                if isinstance(nm, SizeStub):
                    nm = nm.nm
                elif isinstance(nm, Sym):
                    pass  # nm is already set correctly
                elif isinstance(nm, (int, float)):
                    if len(idxs) > 0:
                        self.err(
                            nm_node,
                            f"cannot index '{nm_node.id}' because "
                            f"it is the constant {nm}",
                        )
                    else:
                        return UAST.Const(nm, self.getsrcinfo(e))
                else:  # could not resolve name to anything
                    self.err(nm_node, f"variable '{nm_node.id}' undefined")

                if is_window:
                    return UAST.WindowExpr(nm, idxs, self.getsrcinfo(e))
                else:
                    return UAST.Read(nm, idxs, self.getsrcinfo(e))

        elif isinstance(e, pyast.Attribute):
            if not isinstance(e.value, pyast.Name):
                self.err(
                    e, "expected configuration reads " "of the form 'config.field'"
                )

            assert isinstance(e.attr, str)

            config_obj = e.value.id
            if not self.is_fragment:
                config_obj = self.eval_expr(e.value)
                if not isinstance(config_obj, Config):
                    self.err(e.value, "expected indexed object " "to be a Config")

            return self.AST.ReadConfig(config_obj, e.attr, self.getsrcinfo(e))

        elif isinstance(e, pyast.Constant):
            return self.AST.Const(e.value, self.getsrcinfo(e))

        elif isinstance(e, pyast.UnaryOp):
            if isinstance(e.op, pyast.USub):
                arg = self.parse_expr(e.operand)
                return self.AST.USub(arg, self.getsrcinfo(e))
            else:
                opnm = (
                    "+"
                    if isinstance(e.op, pyast.UAdd)
                    else "not"
                    if isinstance(e.op, pyast.Not)
                    else "~"
                    if isinstance(e.op, pyast.Invert)
                    else "ERROR-BAD-OP-CASE"
                )
                self.err(e, f"unsupported unary operator: {opnm}")

        elif isinstance(e, pyast.BinOp):
            lhs = self.parse_expr(e.left)
            rhs = self.parse_expr(e.right)
            if isinstance(e.op, pyast.Add):
                op = "+"
            elif isinstance(e.op, pyast.Sub):
                op = "-"
            elif isinstance(e.op, pyast.Mult):
                op = "*"
            elif isinstance(e.op, pyast.Div):
                op = "/"
            elif isinstance(e.op, pyast.FloorDiv):
                op = "//"
            elif isinstance(e.op, pyast.Mod):
                op = "%"
            elif isinstance(e.op, pyast.Pow):
                op = "**"
            elif isinstance(e.op, pyast.LShift):
                op = "<<"
            elif isinstance(e.op, pyast.RShift):
                op = ">>"
            elif isinstance(e.op, pyast.BitOr):
                op = "|"
            elif isinstance(e.op, pyast.BitXor):
                op = "^"
            elif isinstance(e.op, pyast.BitAnd):
                op = "&"
            elif isinstance(e.op, pyast.MatMult):
                op = "@"
            else:
                assert False, "unrecognized op"

            if op not in front_ops:
                self.err(e, f"unsupported binary operator: {op}")

            return self.AST.BinOp(op, lhs, rhs, self.getsrcinfo(e))

        elif isinstance(e, pyast.BoolOp):
            assert len(e.values) > 1
            lhs = self.parse_expr(e.values[0])

            if isinstance(e.op, pyast.And):
                op = "and"
            elif isinstance(e.op, pyast.Or):
                op = "or"
            else:
                assert False, "unrecognized op"

            for rhs in e.values[1:]:
                lhs = self.AST.BinOp(op, lhs, self.parse_expr(rhs), self.getsrcinfo(e))

            return lhs

        elif isinstance(e, pyast.Compare):
            assert len(e.ops) == len(e.comparators)

            vals = [self.parse_expr(e.left)] + [
                self.parse_expr(v) for v in e.comparators
            ]
            srcinfo = self.getsrcinfo(e)

            res = None
            for opnode, lhs, rhs in zip(e.ops, vals[:-1], vals[1:]):
                if isinstance(opnode, pyast.Eq):
                    op = "=="
                elif isinstance(opnode, pyast.NotEq):
                    op = "!="
                elif isinstance(opnode, pyast.Lt):
                    op = "<"
                elif isinstance(opnode, pyast.LtE):
                    op = "<="
                elif isinstance(opnode, pyast.Gt):
                    op = ">"
                elif isinstance(opnode, pyast.GtE):
                    op = ">="
                elif isinstance(opnode, pyast.Is):
                    op = "is"
                elif isinstance(opnode, pyast.IsNot):
                    op = "is not"
                elif isinstance(opnode, pyast.In):
                    op = "in"
                elif isinstance(opnode, pyast.NotIn):
                    op = "not in"
                else:
                    assert False, "unrecognized op"
                if op not in front_ops:
                    self.err(e, f"unsupported binary operator: {op}")

                c = self.AST.BinOp(op, lhs, rhs, self.getsrcinfo(e))
                res = c if res is None else self.AST.BinOp("and", res, c, srcinfo)

            return res

        elif isinstance(e, pyast.Call):
            # handle stride expression
            if isinstance(e.func, pyast.Name) and e.func.id == "stride":
                if (
                    len(e.keywords) > 0
                    or len(e.args) != 2
                    or not isinstance(e.args[0], pyast.Name)
                    or not (
                        isinstance(e.args[1], pyast.Constant)
                        and isinstance(e.args[1].value, int)
                        or isinstance(e.args[1], pyast.Name)
                        and e.args[1].id == "_"
                    )
                ):
                    self.err(
                        e,
                        "expected stride(...) to "
                        "have exactly 2 arguments: the identifier "
                        "for the buffer we are talking about "
                        "and an integer specifying which dimension",
                    )

                name = e.args[0].id

                if isinstance(e.args[1], pyast.Name):
                    return PAST.StrideExpr(name, None, self.getsrcinfo(e))

                dim = int(e.args[1].value)
                if not self.is_fragment:
                    if name not in self.locals:
                        self.err(e.args[0], f"variable '{name}' undefined")
                    name = self.locals[name]

                return self.AST.StrideExpr(name, dim, self.getsrcinfo(e))

            # handle built-in functions
            else:
                f = self.eval_expr(e.func)
                fname = e.func.id

                if not isinstance(f, BuiltIn):
                    self.err(
                        e.func, f"expected called object " "to be a builtin function"
                    )

                if len(e.keywords) > 0:
                    self.err(
                        f, "cannot call a builtin function " "with keyword arguments"
                    )
                args = [self.parse_expr(a) for a in e.args]

                return self.AST.BuiltIn(f, args, self.getsrcinfo(e))

        else:
            self.err(e, "unsupported form of expression")



from .LoopIR import LoopIR, T, LoopIR_Rewrite

# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Memory Analysis Pass


class WindowAnalysis(LoopIR_Rewrite):
    def map_s(self, s):
        if isinstance(s, LoopIR.Call):
            args = s.args

            def promote_tensor(a, sa):
                assert sa.type.is_win()
                assert not a.type.is_win()
                assert isinstance(a, LoopIR.Read) and len(a.idx) == 0
                shape = a.type.shape()
                assert len(shape) > 0
                idx = [
                    LoopIR.Interval(LoopIR.Const(0, T.int, N.srcinfo), N, N.srcinfo)
                    for N in shape
                ]
                as_tens = T.Tensor(shape, True, a.type.basetype())
                win_typ = T.Window(a.type, as_tens, a.name, idx)
                return LoopIR.WindowExpr(a.name, idx, win_typ, a.srcinfo)

            def promote_arg(a, sa):
                if sa.type.is_win() and not a.type.is_win():
                    return promote_tensor(a, sa)
                elif (
                    isinstance(sa.type, T.Tensor)
                    and not sa.type.is_win()
                    and a.type.is_win()
                ):
                    raise TypeError(f"{a.srcinfo}: expected a non-window tensor")

                return a

            args = [promote_arg(a, sa) for a, sa in zip(args, s.f.args)]

            return [LoopIR.Call(s.f, args, s.eff, s.srcinfo)]

        return super().map_s(s)

    # make this more efficient by not rewriting
    # most of the sub-trees
    def map_e(self, e):
        return e

    def map_t(self, t):
        return t

    def map_eff(self, eff):
        return eff
